# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# **********************************************  PreferenceSet **********************************************************
#
"""
.. _Preferences:

Preferences specify the behavior of Components when they are executed.  They can be specified by their value, or
by a PreferenceSet, which specifies the

Standard Preferences
--------------------

The following preferences are available for all Components (see `Component prefs <Component_Prefs>` for
additional details):

* **verbosePref** (bool : default False) - enables/disables reporting of (non-exception) warnings and system function;

* **paramValidationPref** (bool : default False) - enables/disables run-time validation of the execute method of a
  Function object;

.. _PreferenceSet_reportOutputPref:

* **reportOutputPref** (`ReportOutput` or list[`ReportOutput`, *PARAMS* or str, str...] : default ReportOutput.OFF) -
  enables/disables and determines format and content for reporting execution of the `Component` (see `ReportOutput`
  for options). The following options can be used (also see `Examples` below):

  * `ReportOutput.OFF` - suppress reporting execution of the Component;

  * `ReportOutput.TERSE` - report only the execution of the Component, but no information about it;

  * `ReportOutput.FULL` - report the input and output of the Component when it is executed;

  * [`ReportOutput.FULL`, *PARAMS*] - report the input and output as well as the value of all of the Component's
  `Parameters` when it is executed;

  * [`ReportOutput.FULL`, <parameter name>, <parameter name>...] - report the value of specific `Parameters` of
      the Component and/or its `function <Component_Function>`.  Specifying *VARIABLE* and/or *VALUE* displays the
      `variable <Mechanism_Base.variable>` and/or `value <Mechanism_Base.value>` attributes of a Mechanism,
      respectivey, and not those attributes of its primary `function <Mechanism_Base.function>` or secondary ones.

  .. note::
     The reportOutputPref settings for a Component only take effect when the Component is executed on it own
     (using its `execute method <Component_Execution>`) or if `ReportOutput.USE_PREFS` is specified in the
     **report_output** argument of one of the Composition's `execution methods <Composition_Execution_Methods>`.
     Otherwise, the ReportOutput option specified in the **report_output** and **report_params** arguments of a
     Composition's execution method takes precedence (see `ReportOutput`).

* **logPref** (`LogCondition` : default LogCondition.OFF) - sets `LogCondition` for a given Component;

COMMENT:
THIS DOES NOT APPEAR CURRENTLY TO BE USED:
* **runtimeParamModulationPref** (Modulation, default: Modulation.MULTIPLY) -  sets type of `Modulation`
  used for modulating parameters by runtime specification (in pathway);
COMMENT

.. technical_note::
   * **deliverPref** (LogCondition, default: LogCondition.OFF) - sets whether attribute data are added to context rpc
     pipeline for delivery to external applications.

.. _Examples:

Examples
--------

reportOutputPref
~~~~~~~~~~~~~~~~

By default, executing a Component, such as the Mechanism below, does not produce any output:

  >>> import psyneulink as pnl
  >>> my_mech = pnl.TransferMechanism()
  >>> my_mech.execute()
  <BLANKLINE>

Output can be specified by specifying a value of `ReportOutput` as the Mechanism's `reportOutputPref` preference.
Assigning `ReportOutput.FULL` generates a report of its input and output values when it was executed:

  >>> my_mech.reportOutputPref = pnl.ReportOutput.FULL
  >>> my_mech.execute()
  ╭─ My Mechanism ─╮
  │ input: 0.0     │
  │ output: 0.0    │
  ╰────────────────╯

Assigning `ReportOutput.TERSE` generates a simpler report:

  >>> my_mech.reportOutputPref = pnl.ReportOutput.TERSE
  >>> my_mech.execute()
   My Mechanism executed

This can be useful when there are many Components executed (e.g., as part of the `execution <Composition_Execution>`
of a complex `Composition`.

Assigning the *PARAMS* keyword with `ReportOutput.FULL` produces a display of the Mechanism's input and output as well
as the value of all of its `Parameters` (for brevity, not all are shown below):

  >>> my_mech.reportOutputPref = [ReportOutput.FULL, pnl.PARAMS]
  >>> my_mech.execute()
  ╭────────────────────────────────────────── My Mechanism ──────────────────────────────────────────╮
  │ input: 0.0                                                                                       │
  │ ╭──────────────────────────────────────────────────────────────────────────────────────────────╮ │
  │ │ params:                                                                                      │ │
     ...
  │ │         integration_rate: 0.5                                                                │ │
     ...
  │ │         noise: 0.0                                                                           │ │
     ...
  │ │         function: Linear Function-6                                                          │ │
  │ │                 intercept: 0.0                                                               │ │
     ...
  │ │                 slope: 1.0                                                                   │ │
  │ │                 value: 0.0                                                                   │ │
  │ │                 variable: 0.0                                                                │ │
  │ │         integrator_function: AdaptiveIntegrator Function-1                                   │ │
     ...
  │ │                 previous_value: 0                                                            │ │
  │ │                 rate: 0.5                                                                    │ │
  │ │                 value: 0.0                                                                   │ │
  │ │                 variable: 0                                                                  │ │
  │ │         termination_measure: Distance Function-5                                             │ │
     ...
  │ │                 metric: max_abs_diff                                                         │ │
  │ │                 normalize: False                                                             │ │
  │ │                 value: 0.0                                                                   │ │
  │ │                 variable: 0 0                                                                │ │
  │ ╰──────────────────────────────────────────────────────────────────────────────────────────────╯ │
  │ output: 0.0                                                                                      │
  ╰──────────────────────────────────────────────────────────────────────────────────────────────────╯

Note that specifying *PARAMS* forces a full output (i.e., equivalent to also specifying `ReportOutput.FULL`).

Generally, not all of a Component's `Parameters` are of interest.  The display can be restricted to
just those of interest by including them in a list specified for reportOutputPref:

  >>> my_mech.reportOutputPref = [ReportOutout.FULL, 'integration_rate', 'slope', 'rate']
  >>> my_mech.execute()
  ╭────────────────────────────────────────── My Mechanism ──────────────────────────────────────────╮
  │ input: 0.0                                                                                       │
  │ ╭──────────────────────────────────────────────────────────────────────────────────────────────╮ │
  │ │ params:                                                                                      │ │
  │ │         integration_rate: 0.5                                                                │ │
  │ │         function: Linear Function-6                                                          │ │
  │ │                 slope: 1.0                                                                   │ │
  │ │         integrator_function: AdaptiveIntegrator Function-1                                   │ │
  │ │                 rate: 0.5                                                                    │ │
  │ ╰──────────────────────────────────────────────────────────────────────────────────────────────╯ │
  │ output: 0.0                                                                                      │
  ╰──────────────────────────────────────────────────────────────────────────────────────────────────╯

Note that this is overridden if `ReportOutput.TERSE` is specified:

  >>> my_mech.reportOutputPref = [pnl.ReportOutput.TERSE, 'integration_rate', 'slope', 'rate']
  >>> my_mech.execute()
   My Mechanism executed

When a Mechanism is executed as part of a Composition, the Composition's reportOutputPref takes precedence:

  >>> my_mech.reportOutputPref = ['integration_rate', 'slope', 'rate']
  >>> my_comp = pnl.Composition(pathways=[my_mech])
  >>> my_comp.run()
  <BLANKLINE>

Note that the Composition's setting, which is ReportOutput.OFF by default, overrode the Mechanism's
reportOutputPref settings.  The **report_output** argument of a Composition's `execution method
<Composition_Execution_Methods>` can also be used to specify report contents;  this too overrides
the Mechanism's reportOutputPref setting:

  >>> my_mech.reportOutputPref = ['integration_rate', 'slope', 'rate', pnl.ReportOutput.FULL]
  >>> my_comp.run(report_output=pnl.ReportOutput.TERSE)
  Composition-0 TRIAL 0 ====================
   Time Step 0 ---------
     My Mechanism executed

`ReportOutput.USE_PREFS` can be used in the **report_output** argument to allow the reportOutputPref of individual
Components to determine what is displayed when a `Composition` is executed.  See `ReportOutput` for other options
and additional examples of how to **report_output** argument of a Composition's `execution methods
<Composition_Execution_Methods>` can be used to configure the output.
"""

import abc
import inspect
from collections import namedtuple
from enum import Enum, IntEnum

from psyneulink.core.globals.keywords import DEFAULT_PREFERENCE_SET_OWNER, PREFERENCE_SET_NAME
from psyneulink.core.globals.utilities import iscompatible, kwCompatibilityType

__all__ = [
    'PreferenceEntry', 'PreferenceLevel', 'PreferenceSet', 'PreferenceSetError', 'PreferenceSetRegistry',
    'PreferenceSetVerbosity'
]

PreferenceSetRegistry = {}

PreferenceSetVerbosity = False

PreferenceEntry = namedtuple('PreferenceEntry', 'setting, level')


class PreferenceLevel(IntEnum):
    NONE        = 0
    INSTANCE    = 1
    SUBTYPE     = 2
    TYPE        = 3
    CATEGORY    = 4
    COMPOSITION = 5


class PreferenceSetError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


@abc.abstractmethod
class PreferenceSet(object):
    """Abstract class for PreferenceSets that stores preferences and provides access to level-specific settings

    COMMENT:
    Description:
        Each PreferenceSet object stores a set of preferences in its corresponding attributes
        Every class in the Component hierarchy is assigned a PreferenceLevel:
            - System:  reserved for the Component class
            - Category: primary function subclasses (e.g., Process, Mechanism, Port, Projection, Function)
            - Type: Category subclasses (e.g., MappingProjection and ControlProjection subclasses of Projection, Function subclasses)
            - Instance: an instance of an object of any class
        Each class level in a hierarchy should be assigned a PreferenceSet object as a class attribute,
            that specifies default settings at that class-level for objects in that class and its subclasses
        Each attribute of a PreferenceSet object is a PreferenceEntry (setting, level) tuple for a preference:
            - setting (value):
                specifies the setting of the preference for the class or object to which the PreferenceSet belongs
            - level (PreferenceLevel):
                there are four levels used to specify default settings at four levels of the class hierarchy,
                    SYSTEM, CATEGORY, TYPE, and INSTANCE
                specifying a given level in PreferenceEntry will cause the value assigned at that level
                    to be returned when a request is made for the value of the setting for that preference
        PreferenceSets are instantiated using a specification dict;  for each entry:
            the key must be a keyPath for a preference attribute (kpReportOutput, kpLog, kpVerbose, kpParamValidation)
            the value must be either:
                a PreferenceEntry, or
                a value that is valid for the setting of the corresponding attribute, or
                a PreferenceLevel
        PreferenceSet attributes MUST have "_pref" as a substring in their attribute name,
            as this is used by the PreferenceSet.show() method (and possibly others in the future)
            to identify PreferenceSet preference attributes
        Any preference attributes defined by a subclass (in its defaultPreferencesDict - see below), but not specified
            in the dict of the prefs arg, will be assigned a default PreferenceEntry from defaultPreferencesDict dict
        PreferenceSet.logPref settings must be assigned a value from a LogEntry class:
            - Globals.Log defines a LogEntry class that can be used
            - classes can define their own LogEntry class;  however:
                every definition of LogEntry must include all of the attributes of Globals.Log.LogEntry
                those attributes must have a value identical to the one in Globals.Log.LogEntry,
                    with the exception of ALL and DEFAULTS (which can vary for different versions of LogEntry)
                assignments from a LogEntry class other than the one in Globals.Log
                    MUST be from one declared in the same module as the object that owns the PreferenceSet
        SUBCLASSES of PreferenceSet:
        Every subclass of PreferenceSet must define a class attribute called defaultPreferencesDict:
            - this must be a dict, with an entry for each preference attribute used by subclass; each entry must have:
                a key that is the name of the instance variable for the corresponding preference attribute
                a value that is a PreferenceEntry specifying the default setting and level for that preference
            - this will be used if no prefs arg is provided to PreferenceSet.__init__, or if any preferences are missing
        Every subclass of PreferenceSet must define a class attribute called baseClass:
            - this must be a class, that is at the base of the hierarchy for which the PreferenceSet will be used
            - this is used to validate PreferenceSet assignments to other classes and objects in the class hierarchy
              and when searching the hierarchy in get_pref_setting_for_level

        [TBI: Deprecated PreferenceSet subclasses; not useful, too complicated;  only implemented MechanismPreferenceSet
            Subclasses of PreferenceSet can be created to define preferences specific to subclasses of Function
                - subclasses of PreferenceSet inherit all the preference attributes above it in the hierarchy
                - however, the level assigned in a PreferenceEntry cannot exceed the level in the PreferenceSet hiearchy
                    at which the preference is defined;  this is checked and, if the level being assigned is too high,
                    a warning message is generated (irrespective of any verbose preferences),
                    and the highest level allowable (i.e., the one at which the preference is defined) is used instead]
            Preference attributes added to a subclass of PreferenceSet MUST have "_pref" as a substring in their name,
                as this is used by the PreferenceSet.show() method (and possibly others in the future)
                to identify PreferenceSet preference attributes

    Initialization arguments:
        - owner (Function object):
            used on instantiation to determine the level of the default preferences;
            assigned dynamically (by the baseClass) on access to a preference;
                this is because a PreferenceSet can be assigned to multiple objects,
                so current owner should be used to determine context-appropriate PreferenceLevel for returning setting
        - level (PreferenceLevel): level for which to report setting (default: PreferenceLevel.SYSTEM)
        - prefs (dict): see above for keys and entries
        - name (str): name of PreferenceSet (default: <subclass.__name__>
        - context (class): must be a subclass of PreferenceSet;  otherwise, an exception is raised
        Note:
        * instantiation always returns a complete PreferenceSet (as defined by the subclass)

    Class attributes:
        - prefsList (list): list of preference attribute names in PreferenceSet (== PreferenceSet.__dict__.keys();

    Class methods:
        - add_preference(pref_attrib=<str>, pref_spec=<value>, default_pref=<PreferenceEntry>):
            create preference attribute for PreferenceSet and assign pref_spec as its value
        - get_pref_setting_for_level(pref_ivar_name=<str>, level=<PreferenceLevel>):
            return setting for specified preference at level specified
            if level is omitted, return setting for level specified in instance's PreferenceEntry
        - show():
            generate table showing all preference attributes for the PreferenceSet, their base and current and values,
                and their PreferenceLevel assignment
        - set_preference(candidate_info=<PreferenceEntry, setting or PreferenceLevel>,
                          default_entry=<PreferenceEntry>,
                          setting_types=<[types]>
                          pref_ivar_name=<str>):
            set specified value of PreferenceEntry attribute (entire PreferenceEntry. setting or level)
        - validate_setting(candidate_setting=<value>, reference_setting=<value>, pref_ivar_name=<str>)
            validate value of setting for pref_ivar_name attribute of PreferenceSet
        - validate_log(candidate_log_item=<LogEntry.<value>>, pref_set=<PreferenceSet>)
            validate that value of kplogPref (_log_pref_) is from conforming and appropriate LogEntry class

    Instance attributes:
        None

    Instance methods:
        None
    COMMENT
    """

    def __init__(self,
                 owner,
                 level=PreferenceLevel.COMPOSITION,
                 prefs=None,
                 name=None,
                 context=None
                 ):
        """Instantiate PreferenceSet from subclass for object and/or class

        --------------------------  CONDITIONS AND ACTIONS:----------------------------------------------------

        Condition    Type    classPref's exist   prefs arg    Action

            1        class		    YES		        NO	     assign self to classPreferences
            2        class	      	YES             dict	 override classPreferences with any prefs
            3        class		    NO		        NO       instantiate ClassPreferences from <subclass>.defaultPrefs
            4        class		    NO		        dict	 instantiate ClassPrefs from defaultPrefs, override w/ prefs

            5        object		    YES		        NO	     use classPreferences
            6        object		    YES		        dict	 use classPreferences as base; override with any prefs
            6.5      object		    YES	(dict)	    NO       instantiate ClassPrefs from defaultPrefs, override w/ dict
            7        object		    NO	            NO   |_| raise exception: these should never occur as instantiation
            8        object		    NO		        dict | |    of object should force instantiation of classPreferences
          # 7        object		    NO	            NO   	 instantiate and use classPreferences from defaultPrefs
          # 8        object		    NO		        dict	 inst. & use classPrefs from defaultPrefs, override w/ prefs
        -------------------------------------------------------------------------------------------------------

        :param owner:
        :param level:
        :param prefs:
        :param name:
        :param context:
        """

        # VALIDATE ATTRIBUTES AND ARGS
        # Make sure subclass implements a baseClass class attribute and that owner is a subclass or instance of it
        try:
            base_class_NOT_A_CLASS = not inspect.isclass(self.baseClass)
        except:
            raise PreferenceSetError("{0} must implement baseClass as a class attribute".
                                     format(self.__class__.__name__))
        else:
            if base_class_NOT_A_CLASS:
                raise PreferenceSetError("{0}.baseClass ({0}) must be a class".
                                         format(self.__class__.__name__, self.baseClass))
            else:
                # owner of PreferenceSet must be a subclass or instance of a subclass in baseClass
                if not (inspect.isclass(owner) or isinstance(owner, self.baseClass)) or owner is None:
                    raise PreferenceSetError("owner argument must be included in call to {1}() "
                                             "and must be an object in the {2} class hierarchy".
                                         format(owner, self.__class__.__name__, self.baseClass.__name__))
                self.owner = owner

        # Make sure subclass implements a defaultPreferencesDict class attribute
        try:
            default_prefs_dict = self.defaultPreferencesDict
        except AttributeError:
            raise PreferenceSetError("{0} must implement defaultPreferencesDict dict as a class attribute".
                                     format(self.__class__.__name__))

        # prefs must be a specification dict or None
        # FIX: replace with typecheck
        if not (isinstance(prefs, dict) or prefs is None):
            raise PreferenceSetError("Preferences ({0}) specified for {1} must a PreferenceSet or"
                                     " specification dict of preferences".format(prefs, owner.name))

        # ASSIGN NAME
        # ****** FIX: 9/10/16: MOVE TO REGISTRY **********

        # FIX: MAKE SURE DEFAULT NAMING SCHEME WORKS WITH CLASSES - 5/30/16
        # FIX: INTEGRATE WITH NAME FROM PREFERENCE_SET_NAME ENTRY IN DICT BELOW - 5/30/16

        if not name:
            # Assign name of preference set class as base of name
            name = self.__class__.__name__
            # If it belongs to a class, append name of owner's class to name
            if inspect.isclass(owner):
                name = name + 'DefaultsFor' + owner.__name__
            # Otherwise, it belongs to an object, so append name of the owner object's class to name
            else:
                name = name + 'Defaultsfor' + owner.__class__.__name__

        # REGISTER
        # FIX: MAKE SURE THIS MAKES SENSE

        from psyneulink.core.globals.registry import register_category
        register_category(entry=self,
                          base_class=PreferenceSet,
                          name=name,
                          registry=PreferenceSetRegistry,
                          )

        # ASSIGN PREFS
        condition = 0

        # Get class preferences (if any) from owner's class
        try:
            class_prefs = owner.classPreferences
        except AttributeError:
            class_prefs = None
        else:
            # Class preferences must be a PreferenceSet or a dict
            if not (isinstance(class_prefs, (PreferenceSet, dict))):
                raise PreferenceSetError("Class preferences for {0} ({1}) must be a PreferenceSet "
                                         "or specification dict for one".format(owner.__name__, class_prefs))

        # Owner is a class
        if inspect.isclass(owner):

            # Class PreferenceSet already exists
            if isinstance(class_prefs, PreferenceSet):
                self = class_prefs

                if prefs is None:                                                    # Condition 1
                    # No prefs arg, so just assign self to class_prefs
                    condition = 1

                else:                                                                          # Condition 2
                    # Get pref from prefs arg or, if not there, from class_prefs
#FIX:  SHOULD TEST FOR *prefsList* ABOVE AND GENERATE IF IT IS NOT THERE, THEN REMOVE TWO SETS OF CODE BELOW THAT DO IT
                    try:
                        self.prefsList
                    except AttributeError:
                        # Generate prefsList for PreferenceSet (since it was not generated when it was created
                        self.prefsList = [i for i in list(class_prefs.__dict__.keys()) if '_pref' in i]

                    for pref_key in self.prefsList:
                        class_pref_entry = getattr(class_prefs, pref_key)
                        try:
                            # Get pref from prefs arg; use class_prefs as validation/default
                            self.set_preference(candidate_info=prefs[pref_key],
                                                default_entry=class_pref_entry,
                                                pref_ivar_name=pref_key)
                        except (KeyError):
                            # No need for default_entry here, as class prefs exist and were validated when instantiated
                            self.set_preference(candidate_info=class_pref_entry,
                                                default_entry=default_prefs_dict[pref_key],  # IMPLEMENTATION: FILLER
                                                pref_ivar_name=pref_key)

                        condition = 2

            # Class preferences either don't exist, or are coming from a specification dict in the class declaration,
            # so use <subclass>.defaultPreferencesDict to instantiate full set, and create PrefsList
            else :
                self.prefsList = []
                if isinstance(class_prefs, dict):
                    # class_prefs are a specification dict from the class declaration
                    # - if prefs are present and are the default set (e.g., in instantiation call from __init__.py),
                    #    replace with class_prefs dict since default_prefs_dict will be used below as base set anyhow)
                    # - if prefs are not the default set, merge with class_prefs dict, giving precedence to prefs
                    #    since they were in a script from the user
                    if prefs is default_prefs_dict:
                        prefs = class_prefs
                    else:
                        prefs.update(class_prefs)
                for pref_key in default_prefs_dict:
                    if pref_key == PREFERENCE_SET_NAME:
                        continue
                    try:
                        # Get pref from prefs arg;                                             # Condition 4
                        #    use default_prefs_dict for validation and backup
                        self.set_preference(candidate_info=prefs[pref_key],
                                            default_entry=default_prefs_dict[pref_key],
                                            pref_ivar_name=pref_key)

                        condition = 4

                    except (KeyError, TypeError) as error:
                        if isinstance(error, KeyError):
                            # Pref not in prefs arg, so get pref from <subclass>.defaultPreferences
                            condition = 4

                        if isinstance(error, TypeError):
                            # No prefs, so get pref from  <subclass>.defaultPreferences        # Condition 3
                            condition = 3
                        # No need for default_entry here, as if the default_prefs_dict is no good there is no hope
                        self.set_preference(candidate_info=default_prefs_dict[pref_key],
                                            default_entry=default_prefs_dict[pref_key],  # IMPLEMENTATION: FILLER
                                            pref_ivar_name=pref_key,)
                        condition = 3

                    # Poplulate prefsList
                    self.prefsList.append(pref_key)

            owner.classPreferences = self

        # Owner is an object
        else:
            if class_prefs:
# # MODIFIED 6/28/16 ADDED:
# # FIX:  class_prefs IS A DICT, SO GETTING AN ERROR BELOW TREATING AS PreferenceSet AND TESTING FOR class_prefs.prefsList
# #       (SEE EXPLANATION ABOVE CONDITION), ??SO ADD THIS:
# # IMPLEMENT Condition 6.5 (see Table above)
#                 if isinstance(class_prefs, dict):
#                     # class_prefs are a specification dict from the class declaration
#                     # - if prefs are present and are the default set (e.g., in instantiation call from __init__.py),
#                     #    replace with class_prefs dict since default_prefs_dict will be used below as base set anyhow)
#                     # - if prefs are not the default set, merge with class_prefs dict, giving precedence to prefs
#                     #    since they were in a script from the user
#                     if prefs is default_prefs_dict:
#                         prefs = class_prefs
#                     else:
#                         prefs.update(class_prefs)
# # MODIFIED 6/28/16 END ADDITION
                try:
                    class_prefs.prefsList
                except AttributeError:
                    # Generate class_prefs.prefsList (since it was not generated when class_prefs was created
                    class_prefs.prefsList = [i for i in list(class_prefs.__dict__.keys()) if '_pref' in i]
                for pref_key in class_prefs.prefsList:
                    try:
                        # Get pref from prefs arg                                              # Condition 6
                        self.set_preference(candidate_info=prefs[pref_key],
                                            default_entry=getattr(class_prefs, pref_key),
                                            pref_ivar_name=pref_key)
                        condition = 6
                    except (KeyError, TypeError) as error:
                        if isinstance(error, KeyError):
                            # Pref not in prefs arg, so get pref from class prefs
                            condition = 6
                        if isinstance(error, TypeError):
                            # No prefs, so get pref from  class_prefs                          # Condition 5
                            condition = 5
                        self.set_preference(candidate_info=getattr(class_prefs, pref_key),
                                            default_entry=default_prefs_dict[pref_key],  # IMPLEMENTATION: FILLER
                                            pref_ivar_name=pref_key)
                owner.prefs = self
            else:                                                                              # Conditions 7 and 8
                raise PreferenceSetError("Program Error: attempt to create a PreferenceSet for object ({0}) "
                                         "in a class ({1}) for which classPreferences have not been instantiated".
                                         format(owner.name, owner.__class__.__name__))

        if PreferenceSetVerbosity:
            print("Preference assignment condition {0}".format(condition))

# FIX: ARE THESE NEEDED?? @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, setting):
        if isinstance(setting, PreferenceLevel):
# FIX: CHECK IF SETTING IS LEGAL FOR THIS LEVEL:
#   TRY CALLING  get_pref_setting_for_level and evaluating return value
#   CHECK FOR ATTRIBUTE ERROR FOR <self>.__class__.classPreferences AT LEVEL SPECIFIED
#   ?? NEED TO TRAVERSE HIERACHY?  ANY WAY TO GO STRAIGHT THERE; OR INCLUDE IN PreferenceLevel STRUCTURE?
#             self.get_pref_setting_for_level(self.name, setting)

            self._level = setting
        else:
            print("Attempt to assign PreferenceSet {0} of {1} an invalid PreferenceLevel ({2});"
                  " level will remain set to {3}".format(self.name, self.owner.name, setting,
                                                         self._level.__class__.__name__ + '.' + self._level.name))

# FIX: @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

#     def add_preference(self, pref_attrib_name, pref_spec, default_entry=NotImplemented):
#         """Instantiate attribute with pref_attrib_name using pref_spec from a preference specification dict
#
#         If pref_spec is for just a setting or just a level, get the other value from default_pref PreferenceEntry
#
#         :param pref_attrib_name: (str)
#         :param pref_spec: (PreferenceEntry, value, or PreferenceLevel
#         :param default_pref: (PreferenceEntry)
#         :return:
#         """
# # FIX: VALIDATE COMPONENTS OF PreferenceEntry HERE (?? ADD METHOD setPreferenceEntry??)
#         # If pref_spec is a PreferenceEntry, use its values to create entry
#         if isinstance(pref_spec, PreferenceEntry):
#             # MODIFIED 6/6/16
#             # setattr(self, pref_attrib_name, pref_spec)
#             self.set_preference(candidate_info=pref_spec, default_entry=default_entry, pref_ivar_name=pref_attrib_name)
#             # setting_OK = self.validate_setting(candidate_info.setting, current_setting, pref_ivar_name)
#             # level_OK = isinstance(candidate_info.level, PreferenceLevel)
#
#         # Otherwise, check if pref_spec is a valid level or setting value
#         else:
#             if isinstance(pref_spec, PreferenceLevel):
#                 # If pref_spec is for a PreferenceLevel, get setting from default_pref
#                 if default_entry is NotImplemented:
#                     raise PreferenceSetError("Program Error: default_pref missing but needed to specify setting for "
#                                              "{0} of {1} in {2}".format(pref_attrib_name,self.name,self.owner.name))
#                 # MODIFIED 6/6/16
#                 # setattr(self, pref_attrib_name, PreferenceEntry(default_pref.setting, pref_spec))
#                 self.set_preference(candidate_info=PreferenceEntry(default_entry.setting, pref_spec),
#                                     default_entry=default_entry,
#                                     pref_ivar_name=pref_attrib_name)
#
#             # If pref_spec is a valid setting (compatible with default_pref value) use it, & get level from default_pref
#             elif default_entry is NotImplemented:
#                 raise PreferenceSetError("Program Error: default_pref missing but needed to specify level for "
#                                          "{0} of {1} in {2}".format(pref_attrib_name,self.name,self.owner.name))
#             # elif iscompatible(pref_spec, default_entry.setting):
#             #     # MODIFIED 6/6/16
#                 # setattr(self, pref_attrib_name, PreferenceEntry(pref_spec, default_pref.level))
#                 self.set_preference(candidate_info=PreferenceEntry(pref_spec, default_entry.level),
#                                     default_entry=default_entry,
#                                     pref_ivar_name=pref_attrib_name)
#             else:
#                 raise PreferenceSetError("{0} is not a valid preference specification for {1};"
#                                          " must be a {2} or PreferenceLevel".
#                                          format(pref_spec, pref_attrib_name, type(pref_spec).__name__))

    def set_preference(self, candidate_info, pref_ivar_name, default_entry=None, skip_validation=False):
        """Validate and assign PreferenceSet, setting, or level to a PreferenceSet preference attribute

        Validate candidate candidate_info and, if OK, assign to pref_ivar_name attribute; candidate_info can be a:
            - PreferenceEntry; validate both the setting and PreferenceLevel items
            - setting; validate against current value retrieved from pref_ivar_name
            - PreferenceLevel; type-check
        If default_entry is provided, it must be a PreferenceEntry;
            use it to validate candidate_info, and fill any invalid or missing values;
        If default_entry is not provided, there must be a current entry for PreferenceSet;
            otherwise an exception will be raised

        Return a PreferenceEntry containing the validated information;
            - this is NOT necessarily the full current PreferenceEntry for the pref;
            - it contains only the information passed in candidate_info (e.g., setting or PreferenceLevel could be None)
            - it is provided simply as parse of candidate_info for the convenience of the caller (e.g., logPref.setter)

        Note: pref_ivar_name MUST include the underscore prefix (e.g., _log_pref), else a recursive loop will ensue
                (wherein the setter calls set_preference, which calls the setter for the pref, etc).

        :param candidate_info: (PreferenceEntry, value, or PreferenceLevel)
        :param pref_ivar_name: (str)
        :param reference_value: ([types])
        :return assignment: (PreferenceEntry)
        """

        # Setup local variables

        # Get owner's name
        if inspect.isclass(self.owner):
            owner_name = self.owner.__name__
        else:
            owner_name = self.owner.name

        # Set all to True, so that if only one is being checked, the other does not interfere with final test
        level_OK = True
        setting_OK = True
        entry_OK = True

        # Get/validate defaults
        # No default specified
        if default_entry is None:
            try:
                # Get current info to use as reference and filler in new PreferenceEntry
                #    in case candidate_info is just a level or setting
                current_entry = getattr(self, pref_ivar_name)
            except AttributeError:
                # # Deal with error if it occurs below, on attempt to inappropriately use default_setting or default_level
                # pass
                raise PreferenceSetError("No existing PreferenceEntry and no default specified for {0}".
                                         format(self.name))
            else:
                default_setting, default_level = current_entry
        # Validate default entry
        else:
            try:
                default_setting, default_level = default_entry
            except TypeError:
                raise PreferenceSetError("Default specification ({0}) for {1} must be a PreferenceEntry".
                                         format(default_entry, self.name))

        # candidate_info is a PreferenceEntry
        if (isinstance(candidate_info, PreferenceEntry)
                or (isinstance(candidate_info, tuple) and len(candidate_info)==2)):
            # elif len(candidate_info) != 2:
            #     raise PreferenceSetError("Preference specification tuple for {} ({}) must have only two entries "
            #                              "(setting and level)".format(owner_name, candidate_info))
            if not isinstance(candidate_info, PreferenceEntry):
                candidate_info = PreferenceEntry(candidate_info[0], candidate_info[1])
            setting_OK = self.validate_setting(candidate_info.setting, default_setting, pref_ivar_name)
            level_OK = isinstance(candidate_info.level, PreferenceLevel)
            if (level_OK and setting_OK) or skip_validation:
                setattr(self, pref_ivar_name, candidate_info)
                return_val = candidate_info
            else:
                entry_OK = False

        # candidate_info is a PreferenceLevel
        elif isinstance(candidate_info, PreferenceLevel):
            setattr(self, pref_ivar_name, PreferenceEntry(default_setting, candidate_info))
            return_val = PreferenceEntry(setting=None, level=candidate_info)

        # candidate_info is a presumed setting
        else:
            setting_OK = self.validate_setting(candidate_info, default_setting, pref_ivar_name)
            if setting_OK or skip_validation:
                setattr(self, pref_ivar_name, PreferenceEntry(candidate_info, default_level))
                return_val = PreferenceEntry(setting=candidate_info, level=None)

        # All is OK, so return
        if (level_OK and setting_OK) or skip_validation:
            return return_val

        # Something's amiss, so raise exception
        if not entry_OK:

            if not level_OK and not setting_OK:
                raise PreferenceSetError("'{0}' and '{1}' are not valid values for setting and PreferenceLevel of"
                                         " PreferenceEntry for {2} in {3} of {4}".
                                         format(candidate_info.setting, candidate_info.level,
                                                pref_ivar_name, self.name, owner_name))
            elif not (level_OK):
                raise PreferenceSetError("'{0}' is not a valid value for PreferenceLevel of PreferenceEntry "
                                         "for {1} in {2} of {3}".
                                         format(candidate_info.level, pref_ivar_name, self.name, owner_name))
            else:
                raise PreferenceSetError("'{0}' is not valid value for setting of PreferenceEntry for {1} in {2}".
                                         format(candidate_info.setting, pref_ivar_name, self.name))

        elif not (setting_OK):
            raise PreferenceSetError("'{0}' is not a valid value for setting of {1} in {2}".
                                     format(str(candidate_info), pref_ivar_name, self.name))
        else:
            raise PreferenceSetError("PROGRAM ERROR")

    def validate_setting(self, candidate_setting, reference_setting, pref_ivar_name):
        """Validate candidate_setting by checking against reference_setting and, if a log_entry, its type

        :param candidate_setting:
        :param reference_setting:
        :return:
        """
        # from Globals.Preferences.BasePreferenceSet import LOG_PREF
        # if pref_ivar_name is LOG_PREF:
        #     self.validate_log(candidate_setting, self)

        setting_OK = (
            iscompatible(candidate_setting, reference_setting, **{kwCompatibilityType:Enum})
            or isinstance(candidate_setting, bool)
        )
        # setting_OK = iscompatible(candidate_setting, reference_setting)

        # if not setting_OK and (isinstance(candidate_setting, Enum) or isinstance(reference_setting, Enum)):
        #     if isinstance(candidate_setting, Enum):
        #         raise PreferenceSetError("'{0}' is not a valid value for setting of {1} in {2} of {3}".
        #                                  format(candidate_setting, pref_ivar_name, self.name, owner_name))
        #     else if

        return setting_OK

    def validate_log(self, candidate_log_item, pref_set):
        """Validate value of log setting and that it is from the LogEntry class associated with the owner of the preference

        :param candidate_log_item:
        :param pref_set:
        :return:
        """
        candidate_log_class = type(candidate_log_item)

        from psyneulink.core.globals.log import LogEntry

        # Candidate_log_item must be from a LogEntry declared in the same module as the owner of the preference
        #    unless it is for the DefaultPeferenceSetOwner (which is used by many classes),
        # Note:
        # * this prevents use of LogEntry attributes not recognized by, or having different values from
        #     the LogEntry class in the owner object's module
        if (not pref_set.owner.name == DEFAULT_PREFERENCE_SET_OWNER and
                candidate_log_class.__module__ is not pref_set.owner.__module__):
            raise PreferenceSetError("Attempt to assign logPref setting for {0} using value ({1}) from LogEntry"
                                     " in {2} which is different than the one defined in Globals.Log"
                                     " or the module ({3}) in which the class of {0} ({4}) was declared".
                                     format(pref_set.owner.name,
                                            candidate_log_item.name,
                                            candidate_log_item.__module__,
                                            pref_set.owner.__module__,
                                            pref_set.owner.__class__.__name__))

        # Validate that every attribute of Globals.Log.LogEntry is present and has the same value in
        #    the LogEntry class used by the setting being assigned
        for log_entry_attribute in LogEntry.__dict__['_member_names_']:

            candidate_log_entry_value = candidate_log_class.__dict__['_member_map_'][log_entry_attribute]
            global_log_entry_value = LogEntry.__dict__['_member_map_'][log_entry_attribute]

            # Validate that attribute from Globals.Log.LogEntry is present in candidate's LogEntry class
            try:
                OK = (candidate_log_entry_value == global_log_entry_value)
            # Attribute from Globals.Log.LogEntry not found in candidate's LogEntry
            except KeyError:
                raise PreferenceSetError("{0} class in {1} must have an attribute {2} identical to the one in {3}".
                                         format(candidate_log_class.__name__,
                                                candidate_log_class.__module__,
                                                log_entry_attribute,
                                                LogEntry.__module__ + "." + LogEntry.__name__,))
            else:
                if OK:
                    continue
                # Skip validation of value of these, as they may differ for different LogEntry classes
                elif (log_entry_attribute == 'ALL' or
                      log_entry_attribute == 'DEFAULTS'):
                    continue
                else:
                    # Value of attribute from candidate's LogEntry does not match the value in Globals.Log.LogEntry
                    raise PreferenceSetError("Value of {0} in {1} ({2}) must have same value as entry in {3} ({4})".
                                             format(log_entry_attribute,
                                                    candidate_log_class.__name__,
                                                    candidate_log_entry_value,
                                                    LogEntry.__module__ + "." + LogEntry.__name__,
                                                    global_log_entry_value))

    def get_pref_setting_for_level(self, pref_ivar_name, requested_level=None):
        """Return the setting of a preference for a specified preference level, and any error messages generated

        Arguments:
        - pref_ivar_name (str): name of ivar for preference attribute for which to return the setting;
        - requested_level (PreferenceLevel): preference level for which the setting should be returned

        Returns:
        - PreferenceEntry.setting, str:
        """
        pref_entry = getattr(self, pref_ivar_name)

        if requested_level is None:
            requested_level = pref_entry.level

        # Preference is owned by an object
        if isinstance(self.owner, self.baseClass):

            # If requested level is INSTANCE, return pref setting for object's PreferenceSet:
            if requested_level is PreferenceLevel.INSTANCE:
                return pref_entry.setting, None

            # If requested level is higher than current one, call class at next level
            elif requested_level > self.owner.__class__.classPreferenceLevel:
                # IMPLEMENTATION NOTE: REMOVE HACK BELOW, ONCE ALL CLASSES ARE ASSIGNED classPreferences ON INIT
                next_level = self.owner.__class__.__bases__[0]
                # If classPreferences for level have not yet been assigned as PreferenceSet, assign them
                if (not hasattr(next_level, 'classPreferences') or
                        not isinstance(next_level.classPreferences, PreferenceSet)):
                    from psyneulink.core.globals.preferences.basepreferenceset import BasePreferenceSet
                    next_level.classPreferences = BasePreferenceSet(owner=next_level,
                                                                    prefs=next_level.classPreferences,
                                                                    level=next_level.classPreferenceLevel)
                return_val = next_level.classPreferences.get_pref_setting_for_level(pref_ivar_name, requested_level)
                return return_val[0],return_val[1]
            # Otherwise, return value for current level
            else:
                # This returns setting from the current level:
                try:
                    return getattr(self.owner.__class__.classPreferences, pref_ivar_name).setting, None
                except AttributeError:
                    # # IMPLEMENTATION NOTE:
                    # #  DEPRECATED THIS, AS CONDITION SHOULD ONLY OCCUR IF
                    # #  A <class>.classPreferences DECLARATION IS MISSING FROM __init__.py
                    # This returns setting from the next level up:
                    # next_level = getattr(self.owner.__class__.__bases__[0].classPreferences, pref_ivar_name).level
                    # next_level_val = getattr(self.owner.__class__.__bases__[0].classPreferences, pref_ivar_name).setting
                    # print ("Warning:  {0} not found at {2} level for {1} class; "
                    #        "replaced with \'{3}\' from {4} class at {5} level".
                    #        format(pref_ivar_name,
                    #               self.owner.__class__.__name__,
                    #               requested_level.name,
                    #               next_level_val,
                    #               self.owner.__class__.__bases__[0].__name__,
                    #               next_level.name))
                    # return next_level_val, None
                    raise PreferenceSetError("Assignment of {}.classPreferences appears to be missing from __init__.py".
                                             format(self.owner.__class__.__name__))

        # Preference is owned by a class
        elif inspect.isclass(self.owner) and issubclass(self.owner, self.baseClass):
            # If requested level is higher than current one:
            if requested_level > self.owner.classPreferenceLevel:
                # Call class at next level

                from psyneulink.core.components.component import Component
                # THis is needed to skip ShellClass, which has no classPreferences, to get to Function (System) level
                if 'ShellClass' in repr(self.owner.__bases__[0]):
                    try:
                        # Store current class pref set (in case class at next level doesn't have the preference)
                        self.previous_level_pref_set = self.owner.classPreferences
                        return_val = Component.classPreferences.get_pref_setting_for_level(pref_ivar_name,
                                                                                    requested_level)
                        return return_val[0], return_val[1]
                    # Pref not found at current level, so use pref from previous level (and report error)
                    except AttributeError:
                        pref_value = getattr(self.previous_level_pref_set, pref_ivar_name).setting
# FIX:  replace level setting??
#                         from Utilities import get_modulationOperation_name
                        err_msg = ("{0} not found at {1}; replaced with value ({2}) from next lower level ({3})".
                                   format(pref_ivar_name,
                                          requested_level.__class__.__name__ + '.' + requested_level.name,
                                          # get_modulationOperation_name(pref_value),
                                          pref_value,
                                          pref_entry.level.__class__.__name__ + '.' + pref_entry.level.name))
                        return pref_value, err_msg
                else:
                    # If classPreferences for level have not yet been assigned as PreferenceSet, assign them
                    next_level = self.owner.__bases__[0]
                    if (not hasattr(next_level, 'classPreferences') or
                            not isinstance(next_level.classPreferences, PreferenceSet)):
                        from psyneulink.core.globals.preferences.basepreferenceset import BasePreferenceSet
                        next_level.classPreferences = BasePreferenceSet(owner=next_level,
                                                                        prefs=next_level.classPreferences,
                                                                        level=next_level.classPreferenceLevel)
                    return_val = self.owner.__bases__[0].classPreferences.get_pref_setting_for_level(pref_ivar_name,
                                                                                               requested_level)
                    return return_val[0], return_val[1]

            # Otherwise, return value for current level
            try:
                return getattr(self, pref_ivar_name).setting, None
            except AttributeError:
                raise PreferenceSetError("PROGRAM ERROR")

        else:
            raise PreferenceSetError("Owner ({0}) of {1} (for which {2} setting is being requested) "
                               "is not a Function object or subclass".
                               format(self.owner.__class__.__name__, self.__class__.__name__, pref_ivar_name))

    def show(self, type=None):
        """Print preferences for PreferenceSet

        :return:
        """

        if self.owner is None:
            raise PreferenceSetError("{0} does not have an owner assigned.  If this PreferenceSet belongs to a class,"
                                     "the class should be given as the owner arg "
                                     "in the call to instantiate the PreferenceSet".format(self.name))

        error_messages = []
        pref_info_table = ""
        # Sort for consistency of reporting
        pref_names_sorted = sorted(self.__dict__.keys())
        for pref_name in pref_names_sorted:
            if '_pref' in pref_name:

                from psyneulink.core.globals.utilities import get_modulationOperation_name

                # GET TABLE INFO
                # Get base_value of pref
                base_value, level = self.__dict__[pref_name]
                # This is needed because value of Modulation is callable (lambda function)
                if inspect.isfunction(base_value):
                    if 'Modulation' in repr(base_value):
                        base_value = get_modulationOperation_name(base_value)
                # Get current_value of pref
                current_value, msg = self.get_pref_setting_for_level(pref_ivar_name=pref_name, requested_level=level)
                if msg:
                    error_messages.append(msg)
                # Get name of Modulation (the value of which is a lambda function)
                if inspect.isfunction(current_value):
                    if 'Modulation' in repr(current_value):
                        current_value = get_modulationOperation_name(current_value)
                # Get name of any enums
                if isinstance(base_value, (Enum, IntEnum)):
                    base_value_str = str(base_value)
                else:
                    base_value_str = repr(base_value)
                if isinstance(current_value, (Enum, IntEnum)):
                    current_value_str = str(current_value)
                else:
                    current_value_str = repr(current_value)
                pref_info_table = pref_info_table +("- {0} {1} {2} {3}\n".format(pref_name[1:].ljust(35,'.'),
                                                                           base_value_str.strip('"\'').ljust(35,'.'),
                                                                           current_value_str.strip('"\'').ljust(35,'.'),
                                                                           level.__class__.__name__ + '.' + level.name))


        # Header info
        print('\nPreferenceSet name: {0}\nPreferenceSet owner: {1}'.format(self.name,
                                                                           self.owner.name.ljust(15)))
        # Error messages (if any):
        if error_messages:
            print()
            for msg in error_messages:
                print(msg)
        # Table:
        print()
        print('{0}{1}{2}{3}'.
              format('Preference:'.ljust(38),
                     'Base Value:'.ljust(36),
                     'Current Value:'.ljust(36),
                     'Level:'.ljust(35)))
        print(pref_info_table)

    # IMPLEMENT:  ADD reference_value ARG THAT IS USED TO VALIDATE SETTING (INSTEAD OF current_setting)
    #             IMPLEMENT USING **{kwCompatibilityType in call to iscompatible)
    #                       BUT WILL NEED TO IMPLEMENT SUPPORT FOR *LIST* OF TYPES FOR kwCompatibilityType


def _assign_prefs(object, prefs, prefs_class:PreferenceSet):

    if isinstance(prefs, PreferenceSet):
        object.prefs = prefs
        # FIX:  CHECK LEVEL HERE??  OR DOES IT NOT MATTER, AS OWNER WILL BE ASSIGNED DYNAMICALLY??
    # Otherwise, if prefs is a specification dict instantiate it, or if it is None assign defaults
    else:
        object.prefs = prefs_class(owner=object, prefs=prefs)
    try:
        # assign log conditions from preferences
        object.parameters.value.log_condition = object.prefs._log_pref.setting
    except AttributeError:
        pass

    try:
        # assign delivery conditions from preferences
        object.parameters.value.delivery_condition = object.prefs._delivery_pref.setting
    except AttributeError:
        pass



import psyneulink as pnl
import numpy as np
import matplotlib.pyplot as plt
#sample Hebb

FeatureNames=['small','medium','large','red','yellow','blue','circle','rectangle','triangle']

# create a variable that corresponds to the size of our feature space
sizeF = len(FeatureNames)
small_red_circle = [1,0,0,1,0,0,1,0,0]
src = small_red_circle


Hebb_comp = pnl.Composition()

Hebb_mech=pnl.RecurrentTransferMechanism(
    input_shapes=sizeF,
    function=pnl.Linear,
    #integrator_mode = True,
    #integration_rate = 0.5,
    enable_learning = True,
    learning_rate = .1,
    name='Hebb_mech',
    #matrix=pnl.AutoAssociativeProjection,
    auto=0,
    hetero=0
    )

Hebb_comp.add_node(Hebb_mech)

Hebb_comp.execution_id = 1

# Use print_info to show numerical values and vis_info to show graphs of the changing values

def print_info():
    print('\nWeight matrix:\n', Hebb_mech.matrix.base, '\nActivity: ', Hebb_mech.value)


def vis_info():
  ub=1#np.amax(np.amax(np.abs(Hebb_mech.matrix.modulated)))
  lb=-ub
  plt.figure()
  plt.imshow(Hebb_mech.matrix.modulated,cmap='RdBu_r',vmin=lb,vmax=ub)
  plt.title('PNL Hebbian Matrix')
  plt.xticks(np.arange(0,sizeF),FeatureNames,rotation=35)
  plt.yticks(np.arange(0,sizeF),FeatureNames,rotation=35)

  plt.colorbar()
  plt.show()

  plt.figure()
  plt.stem(Hebb_mech.value[0])
  plt.title('Activation from Stimulus with PNL Hebbian Matrix')
  plt.xticks(np.arange(0,sizeF),FeatureNames,rotation=35)
  plt.xlabel('Feature')
  plt.ylabel('Activation')
  plt.show()

inputs_dict = {Hebb_mech:np.array(src)}
out=Hebb_comp.learn(num_trials=5,
      # call_after_trial=vis_info,
      inputs=inputs_dict)

print_info()



import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm

x = np.random.rand()
y = np.random.rand()


@pytest.mark.benchmark(group="Builtins")
@pytest.mark.parametrize("op, args, builtin, result", [
                         (np.exp, (x,), "__pnl_builtin_exp", np.exp(x)),
                         #~900 is the limit after which exp returns inf
                         (np.exp, (900.0,), "__pnl_builtin_exp", np.exp(900.0)),
                         (np.log, (x,), "__pnl_builtin_log", np.log(x)),
                         (np.power, (x,y), "__pnl_builtin_pow", np.power(x, y)),
                         (np.tanh, (x,), "__pnl_builtin_tanh", np.tanh(x)),
                         #~450 is the limit after which exp(2x) used in tanh formula returns inf
                         (np.tanh, (450.0,), "__pnl_builtin_tanh", np.tanh(450)),
                         (lambda x: 1.0 / np.tanh(x), (x,), "__pnl_builtin_coth", 1 / np.tanh(x)),
                         #~450 is the limit after which exp(2x) used in coth formula returns inf
                         (lambda x: 1.0 / np.tanh(x), (450,), "__pnl_builtin_coth", 1 / np.tanh(450)),
                         (lambda x: 1.0 / np.sinh(x), (x,), "__pnl_builtin_csch", 1 / np.sinh(x)),
                         #~450 is the limit after which exp(2x) used in csch formula returns inf
                         (lambda x: 1.0 / np.sinh(x), (450,), "__pnl_builtin_csch", 1 / np.sinh(450)),
                         #~900 is the limit after which exp(x) used in csch formula returns inf
                         (lambda x: 1.0 / np.sinh(x), (900,), "__pnl_builtin_csch", 1 / np.sinh(900)),
                         (np.sin, (x,), "__pnl_builtin_sin", np.sin(x)),
                         (np.cos, (x,), "__pnl_builtin_cos", np.cos(x)),
                         ], ids=["EXP", "Large EXP", "LOG", "POW", "TANH", "Large TANH", "COTH", "Large COTH",
                                "CSCH", "Large CSCH", "xLarge CSCH", "SIN", "COS"])
def test_builtin_op(benchmark, op, args, builtin, result, func_mode):
    if func_mode == 'Python':
        f = op

    elif func_mode == 'LLVM':
        f = pnlvm.LLVMBinaryFunction.get(builtin)

    elif func_mode == 'PTX':
        wrap_name = builtin + "_test_wrapper"
        with pnlvm.LLVMBuilderContext.get_current() as ctx:
            intrin = ctx.import_llvm_function(builtin)
            wrap_args = (*intrin.type.pointee.args,
                          intrin.type.pointee.return_type.as_pointer())
            builder = ctx.create_llvm_function(wrap_args, None, wrap_name)
            intrin_args = builder.function.args[:-1]
            ret = builder.call(intrin, intrin_args)
            builder.store(ret, builder.function.args[-1])
            builder.ret_void()

        bin_f = pnlvm.LLVMBinaryFunction.get(wrap_name)

        # The result argument is a pointer, use it to derive
        # the right argument type
        dty = bin_f.np_arg_dtypes[1].base

        ptx_res = np.empty_like(result, dtype=dty)
        ptx_res_arg = pnlvm.jit_engine.pycuda.driver.Out(ptx_res)

        def f(*a):
            bin_f.cuda_call(*(dty.type(p) for p in a), ptx_res_arg)
            return ptx_res

    res = benchmark(f, *args)

    if pytest.helpers.llvm_current_fp_precision() == 'fp32':
        np.testing.assert_allclose(res, result, rtol=1e-5, atol=1e-8)
    else:
        np.testing.assert_allclose(res, result)



import copy
import numpy as np
import psyneulink as pnl
import pytest
import re
import warnings


NO_PARAMETERS = "NO_PARAMETERS"
NO_INIT = "NO_INIT"
NO_VALUE = "NO_VALUE"


def shared_parameter_warning_regex(param_name, shared_name=None):
    if shared_name is None:
        shared_name = param_name

    return (
        f'Specification of the "{param_name}" parameter.*conflicts'
        f' with specification of its shared parameter "{shared_name}"'
    )


# (ancestor, child, should_override)
ancestor_child_data = [
    (pnl.Component, pnl.TransferMechanism, False),
    (pnl.Component, pnl.OutputPort, False),
    (pnl.Component, pnl.InputPort, True),
    (pnl.Component, pnl.SimpleIntegrator, False),
    (pnl.Function_Base, pnl.SimpleIntegrator, True),
    (pnl.TransferMechanism, pnl.RecurrentTransferMechanism, True)
]

# (obj, param_name, alias_name)
param_alias_data = [
    (pnl.Linear, 'slope', 'multiplicative_param'),
    (pnl.Linear, 'intercept', 'additive_param'),
]


@pytest.fixture(scope='function')
def reset_variable(*args):
    yield
    # pytest cannot provide the exact parametrized arguments to fixtures
    # so just reset all of the possibilities
    # this must be used when altering class level defaults
    for item in ancestor_child_data:
        item[0].parameters.variable.reset()
        item[1].parameters.variable.reset()


@pytest.mark.parametrize('ancestor, child', [(item[0], item[1]) for item in ancestor_child_data])
def test_parameter_propagation(ancestor, child):
    for param in ancestor.parameters:
        child_params = child.parameters.values(show_all=True)

        assert param.name in child_params


@pytest.mark.parametrize('ancestor, child, should_override', ancestor_child_data)
def test_parameter_values_overriding(ancestor, child, should_override, reset_variable):
    original_child_variable = child.parameters.variable.default_value

    # ancestor updates
    ancestor.parameters.variable = -1
    assert ancestor.parameters.variable.default_value == -1

    if should_override:
        assert child.parameters.variable.default_value == -1
    else:
        assert child.parameters.variable.default_value == original_child_variable

    # child updates and ancestor does not update
    child.parameters.variable = -2
    assert child.parameters.variable.default_value == -2
    assert ancestor.parameters.variable.default_value == -1

    # child should not get overridden because it is explicitly specified
    ancestor.parameters.variable = -3
    assert child.parameters.variable.default_value == -2

    # revert to original behavior
    child.parameters.variable.reset()
    if should_override:
        assert child.parameters.variable.default_value == -3
    else:
        assert child.parameters.variable.default_value == original_child_variable


def test_unspecified_inheritance():
    class NewTM(pnl.TransferMechanism):
        class Parameters(pnl.TransferMechanism.Parameters):
            pass

    assert NewTM.parameters.variable._inherited
    NewTM.parameters.variable.default_value = -1
    assert not NewTM.parameters.variable._inherited

    NewTM.parameters.variable.reset()
    assert NewTM.parameters.variable._inherited


@pytest.mark.parametrize('obj, param_name, alias_name', param_alias_data)
def test_aliases(obj, param_name, alias_name):
    obj = obj()
    assert obj.parameters._owner is obj
    assert getattr(obj.parameters, alias_name)._owner._owner is obj
    assert getattr(obj.defaults, param_name) == getattr(obj.defaults, alias_name)
    # if hasattr(getattr(obj.parameters, alias_name), 'source'):
    assert getattr(obj.parameters, alias_name).source is getattr(obj.parameters, param_name)


@pytest.mark.parametrize('obj, param_name, alias_name', param_alias_data)
def test_aliases_set_source(obj, param_name, alias_name):
    obj = obj()

    setattr(obj.defaults, param_name, -100)
    assert getattr(obj.defaults, param_name) == getattr(obj.defaults, alias_name)


@pytest.mark.parametrize('obj, param_name, alias_name', param_alias_data)
def test_aliases_set_alias(obj, param_name, alias_name):
    obj = obj()

    setattr(obj.defaults, alias_name, -1)
    assert getattr(obj.defaults, param_name) == getattr(obj.defaults, alias_name)


def test_parameter_getter():
    f = pnl.Linear()
    f.parameters.slope.getter = lambda x: np.asarray(x ** 2)

    assert f.parameters.slope.get(x=3) == 9


def test_parameter_setter():
    f = pnl.Linear()
    f.parameters.slope.setter = lambda x: np.asarray(x ** 2)

    f.parameters.slope.set(3)

    assert f.parameters.slope.get() == 9


def test_history():
    t = pnl.TransferMechanism()
    assert t.parameters.value.get_previous() is None
    t.execute(10)
    assert t.parameters.value.get_previous() == 0
    t.execute(100)
    assert t.parameters.value.get_previous() == 10


@pytest.mark.parametrize(
    'index, range_start, range_end, expected',
    [
        (1, None, None, 4),
        (6, None, None, None),
        (None, 2, None, [3, 4]),
        (None, 2, 0, [3, 4]),
        (1, 2, 0, [3, 4]),
        (None, 5, 2, [0, 1, 2]),
        (None, 10, 2, [0, 1, 2])
    ]
)
def test_get_previous(index, range_start, range_end, expected):
    t = pnl.TransferMechanism()
    t.parameters.value.history_max_length = 10

    for i in range(1, 6):
        t.execute(i)

    previous = t.parameters.value.get_previous(
        index=index,
        range_start=range_start,
        range_end=range_end,
    )

    assert previous == expected


def test_delta():
    t = pnl.TransferMechanism()

    t.execute(10)
    assert t.parameters.value.get_delta() == 10
    t.execute(100)
    assert t.parameters.value.get_delta() == 90


def test_delta_fail():
    t = pnl.TransferMechanism()
    t.parameters.value.set(None, override=True)

    t.execute(10)
    with pytest.raises(TypeError) as error:
        t.parameters.value.get_delta()

    assert "Parameter 'value' value mismatch between current" in str(error.value)


def test_validation():
    class NewTM(pnl.TransferMechanism):
        class Parameters(pnl.TransferMechanism.Parameters):
            variable = pnl.Parameter(np.array([[0], [0], [0]]), read_only=True)

            def _validate_variable(self, variable):
                if not isinstance(variable, np.ndarray) or not variable.shape == np.array([[0], [0], [0]]).shape:
                    return 'must be 2d numpy array of shape (3, 1)'

    t = NewTM()

    t.defaults.variable = np.array([[1], [2], [3]])
    t.parameters.variable.default_value = np.array([[1], [2], [3]])

    with pytest.raises(pnl.ParameterError):
        t.defaults.variable = 0

    with pytest.raises(pnl.ParameterError):
        t.defaults.variable = np.array([0])

    with pytest.raises(pnl.ParameterError):
        t.parameters.variable.default_value = 0

    with pytest.raises(pnl.ParameterError):
        t.parameters.variable.default_value = np.array([[0]])


def test_dot_notation():
    c = pnl.Composition()
    d = pnl.Composition()
    t = pnl.TransferMechanism()
    c.add_node(t)
    d.add_node(t)

    t.execute(1)
    assert t.value == 1
    c.run({t: 5})
    assert t.value == 5
    d.run({t: 10})
    assert t.value == 10
    c.run({t: 20}, context='custom execution id')
    assert t.value == 20

    # context None
    assert t.parameters.value.get() == 1
    assert t.parameters.value.get(c) == 5
    assert t.parameters.value.get(d) == 10
    assert t.parameters.value.get('custom execution id') == 20


def test_copy():
    f = pnl.Linear()
    g = copy.deepcopy(f)

    assert isinstance(g.parameters.additive_param, pnl.ParameterAlias)
    assert g.parameters.additive_param.source is g.parameters.intercept


@pytest.mark.parametrize(
    'cls_, kwargs, parameter, is_user_specified',
    [
        (pnl.AdaptiveIntegrator, {'rate': None}, 'rate', False),
        (pnl.AdaptiveIntegrator, {'rate': None}, 'multiplicative_param', False),
        (pnl.AdaptiveIntegrator, {'rate': 0.5}, 'additive_param', False),
        (pnl.AdaptiveIntegrator, {'rate': 0.5}, 'rate', True),
        (pnl.AdaptiveIntegrator, {'rate': 0.5}, 'multiplicative_param', True),
        (pnl.TransferMechanism, {'integration_rate': None}, 'integration_rate', False),
        (pnl.TransferMechanism, {'integration_rate': 0.5}, 'integration_rate', True),
        (pnl.TransferMechanism, {'initial_value': 0}, 'initial_value', True),
        (pnl.TransferMechanism, {'initial_value': None}, 'initial_value', False),
        (pnl.TransferMechanism, {}, 'initial_value', False),
    ],
)
def test_user_specified(cls_, kwargs, parameter, is_user_specified):
    c = cls_(**kwargs)
    assert getattr(c.parameters, parameter)._user_specified == is_user_specified


@pytest.mark.parametrize(
    'kwargs, parameter, is_user_specified',
    [
        ({'function': pnl.Linear}, 'slope', False),
        ({'function': pnl.Linear()}, 'slope', False),
        ({'function': pnl.Linear(slope=1)}, 'slope', True),
    ]
)
def test_function_user_specified(kwargs, parameter, is_user_specified):
    t = pnl.TransferMechanism(**kwargs)
    assert getattr(t.function.parameters, parameter)._user_specified == is_user_specified


# sort param names or pytest-xdist may cause failure
# see https://github.com/pytest-dev/pytest/issues/4101
@pytest.mark.parametrize('attr', sorted(pnl.Parameter._additional_param_attr_properties))
def test_additional_param_attrs(attr):
    assert hasattr(pnl.Parameter, f'_set_{attr}'), (
        f'To include {attr} in Parameter._additional_param_attr_properties, you'
        f' must add a _set_{attr} method on Parameter. If this is unneeded,'
        ' remove it from Parameter._additional_param_attr_properties.'
    )


class TestSharedParameters:

    recurrent_mech = pnl.RecurrentTransferMechanism(default_variable=[0, 0], enable_learning=True)
    recurrent_mech_no_learning = pnl.RecurrentTransferMechanism(default_variable=[0, 0])
    transfer_with_costs = pnl.TransferWithCosts(default_variable=[0, 0])

    test_values = [
        (
            recurrent_mech,
            'learning_function',
            recurrent_mech.learning_mechanism.parameters.function
        ),
        (
            recurrent_mech,
            'learning_rate',
            recurrent_mech.learning_mechanism.parameters.learning_rate
        ),
        (
            transfer_with_costs,
            'transfer_fct_mult_param',
            transfer_with_costs.transfer_fct.parameters.multiplicative_param
        )
    ]

    @pytest.mark.parametrize(
        'obj, parameter_name, source',
        test_values + [
            (recurrent_mech_no_learning, 'learning_function', None),
        ]
    )
    def test_sources(self, obj, parameter_name, source):
        assert getattr(obj.parameters, parameter_name).source is source

    @pytest.mark.parametrize(
        'obj, parameter_name, source',
        test_values
    )
    def test_values(self, obj, parameter_name, source):
        obj_param = getattr(obj.parameters, parameter_name)
        eids = range(5)

        for eid in eids:
            obj.execute(np.array([eid, eid]), context=eid)

        context = pnl.Context(execution_id=eid)
        assert all([
            obj_param._get(context) is source._get(context)
            for eid in eids
        ])

    @pytest.mark.parametrize(
        'obj, parameter_name, attr_name',
        [
            (transfer_with_costs, 'intensity_cost_fct_mult_param', 'modulable'),
            (recurrent_mech, 'learning_function', 'stateful'),
            (recurrent_mech, 'learning_function', 'loggable'),
            (recurrent_mech.recurrent_projection, 'auto', 'modulable'),
            (recurrent_mech, 'integration_rate', 'modulable'),
            (recurrent_mech, 'noise', 'modulable'),
        ]
    )
    def test_param_attrs_match(self, obj, parameter_name, attr_name):
        shared_param = getattr(obj.parameters, parameter_name)
        source_param = shared_param.source

        assert getattr(shared_param, attr_name) == getattr(source_param, attr_name)

        orig_values = shared_param.stateful, source_param.stateful

        # change value of shared attribute on source parameter
        source_param.stateful = not source_param.stateful
        assert getattr(shared_param, attr_name) == getattr(source_param, attr_name)

        shared_param.stateful, source_param.stateful = orig_values

        # change value of shared attribute on sharedparameter
        shared_param.stateful = not shared_param.stateful
        assert getattr(shared_param, attr_name) == getattr(source_param, attr_name)

        shared_param.stateful, source_param.stateful = orig_values

    @pytest.mark.parametrize(
        'integrator_function, expected_rate',
        [
            (pnl.AdaptiveIntegrator, pnl.TransferMechanism.defaults.integration_rate),
            (pnl.AdaptiveIntegrator(), pnl.TransferMechanism.defaults.integration_rate),
            (pnl.AdaptiveIntegrator(rate=.75), .75)
        ]
    )
    def test_override_tmech(self, integrator_function, expected_rate):
        t = pnl.TransferMechanism(integrator_function=integrator_function)
        assert t.integrator_function.defaults.rate == expected_rate
        assert t.integration_rate.modulated == t.integration_rate.base == expected_rate

    def test_conflict_warning(self):
        with pytest.warns(
            UserWarning,
            match=shared_parameter_warning_regex('integration_rate', 'rate')
        ):
            pnl.TransferMechanism(
                integration_rate=.1,
                integrator_function=pnl.AdaptiveIntegrator(rate=.2)
            )

    @pytest.mark.parametrize(
        'mech_type, param_name, shared_param_name, param_value',
        [
            (pnl.LCAMechanism, 'noise', 'noise', pnl.GaussianDistort),
            (pnl.LCAMechanism, 'noise', 'noise', pnl.GaussianDistort()),
            (pnl.TransferMechanism, 'noise', 'noise', pnl.NormalDist),
            (pnl.TransferMechanism, 'noise', 'noise', pnl.NormalDist()),
            (pnl.TransferMechanism, 'noise', 'noise', [pnl.NormalDist()]),
        ]
    )
    def test_conflict_no_warning(
        self,
        mech_type,
        param_name,
        shared_param_name,
        param_value
    ):
        # pytest doesn't support inverse warning assertion for specific
        # warning only
        with warnings.catch_warnings():
            warnings.simplefilter(action='error', category=UserWarning)
            try:
                mech_type(**{param_name: param_value})
            except UserWarning as w:
                if re.match(shared_parameter_warning_regex(param_name, shared_param_name), str(w)):
                    raise

    def test_conflict_no_warning_parser(self):
        # replace with different class/parameter if _parse_noise ever implemented
        assert not hasattr(pnl.AdaptiveIntegrator.parameters, '_parse_noise')
        pnl.AdaptiveIntegrator.parameters._parse_noise = lambda noise: np.array(2 * noise)

        # pytest doesn't support inverse warning assertion for specific
        # warning only
        with warnings.catch_warnings():
            warnings.simplefilter(action='error', category=UserWarning)
            try:
                pnl.TransferMechanism(
                    noise=2,
                    integrator_function=pnl.AdaptiveIntegrator(noise=1)
                )
            except UserWarning as w:
                if re.match(shared_parameter_warning_regex('noise'), str(w)):
                    raise

        delattr(pnl.AdaptiveIntegrator.parameters, '_parse_noise')


class TestSpecificationType:
    @staticmethod
    def _create_params_class_variant(cls_param, init_param, parent_class=pnl.Component):
        # init_param as Parameter doesn't make sense, only check cls_param
        if cls_param is pnl.Parameter:
            cls_param = pnl.Parameter()

        if cls_param is NO_PARAMETERS:
            if init_param is NO_INIT:

                class TestComponent(parent_class):
                    pass

            else:

                class TestComponent(parent_class):
                    @pnl.core.globals.parameters.check_user_specified
                    def __init__(self, p=init_param):
                        super().__init__(p=p)

        elif cls_param is NO_VALUE:
            if init_param is NO_INIT:

                class TestComponent(parent_class):
                    class Parameters(parent_class.Parameters):
                        pass

            else:

                class TestComponent(parent_class):
                    class Parameters(parent_class.Parameters):
                        pass

                    @pnl.core.globals.parameters.check_user_specified
                    def __init__(self, p=init_param):
                        super().__init__(p=p)

        else:
            if init_param is NO_INIT:

                class TestComponent(parent_class):
                    class Parameters(parent_class.Parameters):
                        p = cls_param

            else:

                class TestComponent(parent_class):
                    class Parameters(parent_class.Parameters):
                        p = cls_param

                    @pnl.core.globals.parameters.check_user_specified
                    def __init__(self, p=init_param):
                        super().__init__(p=p)

        return TestComponent

    @pytest.mark.parametrize(
        "cls_param, init_param, param_default",
        [
            (1, 1, 1),
            (1, None, 1),
            (None, 1, 1),
            (1, NO_INIT, 1),
            ("foo", "foo", "foo"),
            (np.array(1), np.array(1), np.array(1)),
            (np.array([1]), np.array([1]), np.array([1])),
        ],
    )
    def test_valid_assignment(self, cls_param, init_param, param_default):
        TestComponent = TestSpecificationType._create_params_class_variant(cls_param, init_param)
        assert TestComponent.defaults.p == param_default
        assert TestComponent.parameters.p.default_value == param_default

    @pytest.mark.parametrize(
        "cls_param, init_param",
        [
            (1, 2),
            (2, 1),
            (1, 1.0),
            (np.array(1), 1),
            (np.array([1]), 1),
            (np.array([1]), np.array(1)),
            ("foo", "bar"),
        ],
    )
    def test_conflicting_assignments(self, cls_param, init_param):
        with pytest.raises(AssertionError, match="Conflicting default parameter"):
            TestSpecificationType._create_params_class_variant(cls_param, init_param)

    @pytest.mark.parametrize(
        "child_cls_param, child_init_param, parent_value, child_value",
        [
            (NO_PARAMETERS, NO_INIT, 1, 1),
            (NO_VALUE, NO_INIT, 1, 1),
            (2, NO_INIT, 1, 2),
            (NO_PARAMETERS, 2, 1, 2),
            (NO_VALUE, 2, 1, 2),
            (2, 2, 1, 2),
        ],
    )
    @pytest.mark.parametrize(
        "parent_cls_param, parent_init_param",
        [(1, 1), (1, None), (None, 1), (pnl.Parameter, 1)],
    )
    def test_inheritance(
        self,
        parent_cls_param,
        parent_init_param,
        child_cls_param,
        child_init_param,
        parent_value,
        child_value,
    ):
        TestParent = TestSpecificationType._create_params_class_variant(
            parent_cls_param, parent_init_param
        )
        TestChild = TestSpecificationType._create_params_class_variant(
            child_cls_param, child_init_param, parent_class=TestParent
        )

        assert TestParent.defaults.p == parent_value
        assert TestParent.parameters.p.default_value == parent_value

        assert TestChild.defaults.p == child_value
        assert TestChild.parameters.p.default_value == child_value

    @pytest.mark.parametrize("set_from_defaults", [True, False])
    @pytest.mark.parametrize(
        "child_cls_param, child_init_param",
        [(1, 1), (1, None), (None, 1), (NO_PARAMETERS, 1), (1, NO_INIT)],
    )
    @pytest.mark.parametrize("parent_cls_param, parent_init_param", [(0, 0), (0, None)])
    def test_set_and_reset(
        self,
        parent_cls_param,
        parent_init_param,
        child_cls_param,
        child_init_param,
        set_from_defaults,
    ):
        def set_p_default(obj, val):
            if set_from_defaults:
                obj.defaults.p = val
            else:
                obj.parameters.p.default_value = val

        TestParent = TestSpecificationType._create_params_class_variant(
            parent_cls_param, parent_init_param
        )
        TestChild = TestSpecificationType._create_params_class_variant(
            child_cls_param, child_init_param, parent_class=TestParent
        )
        TestGrandchild = TestSpecificationType._create_params_class_variant(
            NO_PARAMETERS, NO_INIT, parent_class=TestChild
        )

        set_p_default(TestChild, 10)
        assert TestParent.defaults.p == 0
        assert TestChild.defaults.p == 10
        assert TestGrandchild.defaults.p == 10

        set_p_default(TestGrandchild, 20)
        assert TestParent.defaults.p == 0
        assert TestChild.defaults.p == 10
        assert TestGrandchild.defaults.p == 20

        TestChild.parameters.p.reset()
        assert TestParent.defaults.p == 0
        assert TestChild.defaults.p == 1
        assert TestGrandchild.defaults.p == 20

        TestGrandchild.parameters.p.reset()
        assert TestParent.defaults.p == 0
        assert TestChild.defaults.p == 1
        assert TestGrandchild.defaults.p == 1

        set_p_default(TestGrandchild, 20)
        assert TestParent.defaults.p == 0
        assert TestChild.defaults.p == 1
        assert TestGrandchild.defaults.p == 20


def test_dependent_parameter_validate():
    # using 3 parameters to reduce chance of random success
    class NewF(pnl.Function_Base):
        class Parameters(pnl.Function_Base.Parameters):
            a = pnl.Parameter(1)
            b = pnl.Parameter(2, dependencies='a')
            c = pnl.Parameter(3, dependencies='b')
            d = pnl.Parameter(4, dependencies='c')

            def _validate_b(self, b):
                if b != self.a.default_value + 1:
                    return 'invalid'

            def _validate_c(self, c):
                if c != self.b.default_value + 1:
                    return 'invalid'

            def _validate_d(self, d):
                if d != self.c.default_value + 1:
                    return 'invalid'

        def __init__(self, **kwargs):
            return super().__init__(0, {}, **kwargs)

        def _function(self, variable=None, context=None, params=None):
            return np.array(0)

    pnl.ProcessingMechanism(function=NewF(a=2, b=3, c=4, d=5))

    with pytest.raises(pnl.ParameterError) as err:
        # b should be first error to occur
        pnl.ProcessingMechanism(function=NewF(b=3, c=5, d=7))
    assert re.match(
        r"Value \(3\) assigned to parameter 'b'.*is not valid: invalid",
        str(err.value)
    )


def test_dependency_from_parameter():
    class NewM(pnl.ProcessingMechanism):
        class Parameters(pnl.ProcessingMechanism.Parameters):
            a = pnl.Parameter(None)
            b = pnl.Parameter(None, dependencies='a')
            c = pnl.Parameter(None, dependencies=b)

    m = NewM()
    param_order = m.parameters._in_dependency_order
    assert param_order.index(m.parameters.c) > param_order.index(m.parameters.b)



import numpy as np
import psyneulink as pnl
import pytest


class TestMechanism:

    @pytest.mark.mechanism
    @pytest.mark.parametrize(
        'mechanism, default_variable, result_variable',
        [
            (pnl.TransferMechanism, [0], np.array([[0]])),
            (pnl.IntegratorMechanism, [0], np.array([[0]])),
        ]
    )
    def test_transfer_mech_instantiation(self, mechanism, default_variable, result_variable):
        T = mechanism(default_variable=default_variable)

        assert T.defaults.variable == result_variable
        assert T.defaults.value == result_variable

        assert T.function.defaults.variable == result_variable
        assert T.function.defaults.value == result_variable

        assert T.input_port.defaults.variable == result_variable[0]
        assert T.input_port.defaults.value == result_variable[0]

        assert T.input_port.function.defaults.variable == result_variable[0]
        assert T.input_port.function.defaults.value == result_variable[0]

    @pytest.mark.mechanism
    @pytest.mark.parametrize(
        'mechanism_type, default_variable, mechanism_value, function_value',
        [
            (pnl.ObjectiveMechanism, [0, 0, 0], np.array([[0, 0, 0]]), np.array([[0, 0, 0]]))
        ]
    )
    def test_value_shapes(self, mechanism_type, default_variable, mechanism_value, function_value):
        M = mechanism_type(default_variable=default_variable)

        assert M.defaults.value.shape == mechanism_value.shape
        assert M.function.defaults.value.shape == function_value.shape

    @pytest.mark.parametrize(
        'noise',
        [pnl.GaussianDistort, pnl.NormalDist]
    )
    def test_noise_assignment_equivalence(self, noise):
        t1 = pnl.TransferMechanism(name='t1', input_shapes=2, noise=noise())
        t2 = pnl.TransferMechanism(name='t2', input_shapes=2)
        t2.integrator_function.parameters.noise.set(noise())

        t1.integrator_function.noise.seed.base = 0
        t2.integrator_function.noise.base.seed = 0

        for _ in range(5):
            np.testing.assert_equal(t1.execute([1, 1]), t2.execute([1, 1]))

    @pytest.mark.parametrize(
        'noise, included_parameter_ports, excluded_parameter_ports, noise_statefulness',
        [
            (0, ['noise'], ['seed'], True),
            ([0], ['noise'], ['seed'], True),
            ([0, 0], ['noise'], ['seed'], True),
            ([0, pnl.NormalDist()], [], ['noise', 'seed'], False),
            (pnl.NormalDist, ['seed'], ['noise'], False),
            ([pnl.NormalDist(), pnl.NormalDist()], [], ['noise', 'seed'], False),
        ]
    )
    def test_numeric_noise_specifications(
        self,
        noise,
        included_parameter_ports,
        excluded_parameter_ports,
        noise_statefulness
    ):
        try:
            size = len(noise)
        except TypeError:
            size = 1

        t = pnl.TransferMechanism(input_shapes=size, noise=noise)

        assert all(p in t.parameter_ports for p in included_parameter_ports)
        assert all(p not in t.parameter_ports for p in excluded_parameter_ports)

        assert t.parameters.noise.stateful is noise_statefulness

    @pytest.mark.parametrize(
        'noise',
        [
            [0, pnl.NormalDist()],
            pnl.NormalDist,
            [pnl.NormalDist(), pnl.NormalDist()]
        ]
    )
    def test_noise_change_warning_to_numeric(self, noise):
        try:
            size = len(noise)
        except TypeError:
            size = 1

        t = pnl.TransferMechanism(input_shapes=size, noise=noise)

        with pytest.warns(
            UserWarning,
            match='Setting noise to a numeric value after instantiation.*'
        ):
            t.parameters.noise.set(0)

    @pytest.mark.parametrize(
        'noise',
        [
            0,
            [0],
            [0, 0],
        ]
    )
    def test_noise_change_warning_to_function(self, noise):
        try:
            size = len(noise)
        except TypeError:
            size = 1

        t = pnl.TransferMechanism(input_shapes=size, noise=noise)

        with pytest.warns(
            UserWarning,
            match='Setting noise to a value containing functions after instantiation.*'
        ):
            t.parameters.noise.set(pnl.NormalDist)

    def test_execute_no_input_doesnt_change_default_variable(self):
        m = pnl.ProcessingMechanism(default_variable=1)
        assert m.defaults.variable == [[1]]
        m.execute()
        assert m.defaults.variable == [[1]]
        m.execute(2)
        assert m.defaults.variable == [[1]]


class TestMechanismFunctionParameters:
    f = pnl.Linear()
    i = pnl.SimpleIntegrator()
    mech_1 = pnl.TransferMechanism(function=f, integrator_function=i)
    mech_2 = pnl.TransferMechanism(function=f, integrator_function=i)
    integrator_mechanism = pnl.IntegratorMechanism(function=i)

    @pytest.mark.parametrize(
        "f, g",
        [
            pytest.param(
                mech_1.defaults.function,
                mech_2.defaults.function,
                id="function_defaults",
            ),
            pytest.param(
                mech_1.defaults.function,
                mech_1.parameters.function.get(),
                id="function_default-and-value",
            ),
            pytest.param(
                mech_1.defaults.function,
                mech_2.parameters.function.get(),
                id="function_default-and-other-value",
            ),
            pytest.param(
                mech_1.defaults.integrator_function,
                mech_2.defaults.integrator_function,
                id="integrator_function_defaults",
            ),
            pytest.param(
                mech_1.defaults.integrator_function,
                mech_1.parameters.integrator_function.get(),
                id="integrator_function_default-and-value",
            ),
            pytest.param(
                mech_1.defaults.integrator_function,
                mech_2.parameters.integrator_function.get(),
                id="integrator_function_default-and-other-value",
            ),
        ],
    )
    def test_function_parameter_distinctness(self, f, g):
        assert f is not g

    @pytest.mark.parametrize(
        "f, owner",
        [
            pytest.param(
                mech_1.parameters.function.get(),
                mech_1,
                id='function'
            ),
            pytest.param(
                integrator_mechanism.class_defaults.function,
                integrator_mechanism.class_parameters.function,
                id="class_default_function"
            ),
            pytest.param(
                mech_1.defaults.function,
                mech_1.parameters.function,
                id="default_function"
            ),
            pytest.param(
                mech_1.parameters.termination_measure.get(),
                mech_1,
                id='termination_measure'
            ),
            pytest.param(
                mech_1.class_defaults.termination_measure,
                mech_1.class_parameters.termination_measure,
                id="class_default_termination_measure"
            ),
            pytest.param(
                mech_1.defaults.termination_measure,
                mech_1.parameters.termination_measure,
                id="default_termination_measure"
            ),
        ]
    )
    def test_function_parameter_ownership(self, f, owner):
        assert f.owner is owner

    @pytest.mark.parametrize(
        'param_name, function',
        [
            ('function', f),
            ('integrator_function', i),
        ]
    )
    def test_function_parameter_assignment(self, param_name, function):
        # mech_1 should use the exact instances, mech_2 should have copies
        assert getattr(self.mech_1.parameters, param_name).get() is function
        assert getattr(self.mech_2.parameters, param_name).get() is not function

class TestResetValues:

    def test_reset_state_integrator_mechanism(self):
        A = pnl.IntegratorMechanism(name='A', function=pnl.DriftDiffusionIntegrator(time_step_size=1.0))

        # Execute A twice
        #  [0] saves decision variable only (not time)
        original_output = [A.execute(1.0), A.execute(1.0)]

        # SAVING STATE  - - - - - - - - - - - - - - - - - - - - - - - - -
        reset_stateful_functions_to = {}
        for attr in A.function.stateful_attributes:
            reset_stateful_functions_to[attr] = getattr(A.function, attr)

        print(reset_stateful_functions_to)
        # Execute A twice AFTER saving the state so that it continues accumulating.
        # We expect the next two outputs to repeat once we reset the state b/c we will return it to the current state
        output_after_saving_state = [A.execute(1.0), A.execute(1.0)]

        # RESETTING STATE - - - - - - - - - - - - - - - - - - - - - - - -
        A.reset(**reset_stateful_functions_to)

        # We expect these results to match the results from immediately after saving the state
        output_after_reinitialization = [A.execute(1.0), A.execute(1.0)]

        np.testing.assert_allclose(output_after_saving_state, output_after_reinitialization)
        np.testing.assert_allclose(original_output, np.array([[[1.0], [1.0]], [[2.0], [2.0]]]))
        np.testing.assert_allclose(output_after_reinitialization, np.array([[[3.0], [3.0]], [[4.0], [4.0]]]))

    def test_reset_state_transfer_mechanism(self):
        A = pnl.TransferMechanism(name='A', integrator_mode=True)

        # Execute A twice
        original_output = [A.execute(1.0), A.execute(1.0)]

        # SAVING STATE  - - - - - - - - - - - - - - - - - - - - - - - - -
        reset_stateful_functions_to = {}

        for attr in A.integrator_function.stateful_attributes:
            reset_stateful_functions_to[attr] = getattr(A.integrator_function, attr)

        # Execute A twice AFTER saving the state so that it continues accumulating.
        # We expect the next two outputs to repeat once we reset the state b/c we will return it to the current state
        output_after_saving_state = [A.execute(1.0), A.execute(1.0)]

        # RESETTING STATE - - - - - - - - - - - - - - - - - - - - - - - -
        A.reset(**reset_stateful_functions_to)

        # We expect these results to match the results from immediately after saving the state
        output_after_reinitialization = [A.execute(1.0), A.execute(1.0)]

        np.testing.assert_allclose(output_after_saving_state, output_after_reinitialization)
        np.testing.assert_allclose(original_output, [np.array([[0.5]]), np.array([[0.75]])])
        np.testing.assert_allclose(output_after_reinitialization, [np.array([[0.875]]), np.array([[0.9375]])])

    @pytest.mark.usefixtures("comp_mode_no_llvm")
    def test_reset_integrator_function(self, comp_mode):
        """This test checks that the Mechanism.integrator_function is reset when the mechanism is"""

        threshold_mech = pnl.TransferMechanism(input_shapes=1,
                                               default_variable=0,
                                               integrator_function=pnl.SimpleIntegrator(rate=1, offset=-0.001),
                                               function=pnl.Linear(intercept=0.06, slope=1),
                                               integrator_mode=True,
                                               execute_until_finished=True,
                                               termination_threshold=10,
                                               reset_stateful_function_when=pnl.AtTrialStart(),
                                               termination_measure=pnl.TimeScale.TRIAL)
        comp = pnl.Composition()
        comp.add_node(threshold_mech)

        results = comp.run(inputs=[[0.0], [0.0]], execution_mode=comp_mode)
        np.testing.assert_allclose(results, [[0.05]])



# Princeton University licenses this file to You under the Apache License,
# Version 2.0 (the "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# ***************************** PredictionErrorMechanism ***********************

"""

Contents
--------

  * `PredictionErrorMechanism_Overview`
  * `PredictionErrorMechanism_Creation`
  * `PredictionErrorMechanism_Structure`
  * `PredictionErrorMechanism_Execution`
  * `PredictionErrorMechanism_Example`
  * `PredictionErrorMechanism_Class_Reference`


.. _PredictionErrorMechanism_Overview:

Overview
--------

A PredictionErrorMechanism is a subclass of `ComparatorMechanism` that receives two inputs (a sample and a target),
and calculates the temporal difference prediction error as found in `Montague, Dayan, and Sejnowski (1996)
<http://www.jneurosci.org/content/jneuro/16/5/1936.full.pdf>`_ using its `function
<PredictionErrorMechanism.function>`, and places the delta values (the difference between the actual and predicted
reward) in its *OUTCOME* `OutputPort <PredictionErrorMechanism.output_port>`.

.. _PredictionErrorMechanism_Creation:

Creating a PredictionErrorMechanism
-----------------------------------

A PredictionErrorMechanism is usually created automatically when a `LearningMechanism`
`is created <LearningMechanism_Creation>` using the `TDLearning` function).
A PredictionErrorMechanism can also be created directly by calling its constructor.
Its **sample** and **target**  arguments are used to specify the OutputPorts
that provide the sample and target inputs, respectively (see
`ObjectiveMechanism Monitored Output Ports <ObjectiveMechanism_Monitor>`
for details). When the PredictionErrorMechanism is created, two InputPorts are
created, one each for its sample and target inputs (and named, by default
*SAMPLE* and *TARGET*). Each is assigned a MappingProjection from the
corresponding OutputPort specified in the **sample** and **target** arguments.

It is important to recognize that the value of the *SAMPLE* and *TARGET*
InputPorts must have the same length and type, so that they can be compared
using the PredictionErrorMechanism's `function
<PredictionErrorMechanism.function>.` By default, they use the format of the
OutputPorts specified in the **sample** and **target** arguments, respectively,
and the `MappingProjection` to each uses an `IDENTITY_MATRIX`. Therefore, for
the default configuration, the OutputPorts specified in the **sample** and
**target** arguments must have values of the same length and type. If these
differ, the **input_ports** argument can be used to explicitly specify the
format of the PredictionErrorMechanism's *SAMPLE* and *TARGET* InputPorts, to
insure they are compatible with one another (as well as to customize their
names, if desired). If the **input_ports** argument is used, *both* the sample
and target InputPorts must be specified. Any of the formats for `specifying
InputPorts <InputPort_Specification>` can be used in the argument. If values
are assigned for the InputPorts, they must be of equal length and type. Their
types must also be compatible with the value of the OutputPorts specified in
the **sample** and **target** arguments. However, the length specified for an
InputPort can differ from its corresponding OutputPort; in that case, by
default, the MappingProjection created uses a `FULL_CONNECTIVITY` matrix. Thus,
OutputPorts of differing lengths can be mapped to the sample and target
InputPorts of a PredictionErrorMechanism (see the `example
<PredictionErrorMechanism_Example>` below), so long as the latter of of the
same length. If a projection other than a `FULL_CONNECTIVITY` matrix is
needed, this can be specified using the *PROJECTION* entry of a `Port
specification dictionary <Port_Specification>` for the InputPort in the
**input_ports** argument.

.. _PredictionErrorMechanism_Structure:

Structure
---------

A PredictionErrorMechanism has two `input_ports
<ComparatorMechanism.input_ports>`, each of which receives a
`MappingProjection` from a corresponding OutputPort specified in the
**sample** and **target** arguments of its constructor. The InputPorts are
listed in the Mechanism's `input_ports <ComparatorMechanism.input_ports>`
attribute and named, respectively, *SAMPLE* and *TARGET*. The OutputPorts
from which they receive their projections (specified in the **sample** and
**target** arguments) are listed in the Mechanism's `sample
<ComparatorMechanism.sample>` and `target
<ComparatorMechanism.target>` attributes as well as in its
`monitored_output_ports <ComparatorMechanism.monitored_output_ports>`
attribute. The PredictionErrorMechanism's `function
<PredictionErrorMechanism.function>` calculates the difference between the
predicted reward and the true reward at each timestep in **SAMPLE**. By
default, it uses a `PredictionErrorDeltaFunction`. However, the
`function <PredictionErrorMechanism.function>` can be customized, so long as it
is replaced with one that takes two arrays with the same format as its inputs
and generates a similar array as its result. The result is assigned as the
value of the PredictionErrorMechanism's *OUTCOME* (`primary
<OutputPort_Primary>`) OutputPort.

.. _PredictionErrorMechanism_Execution:

Execution
---------

When a PredictionErrorMechanism is executed, it updates its input_ports with
the values of the OutputPorts specified in its **sample** and **target**
arguments, and then uses its `function <PredictionErrorMechanism.function>` to
compare these. By default, the result is assigned to the `value
<PredictionErrorMechanism.value>` of its *OUTCOME* `output_port
<PredictionErrorMechanism.output_port>`, and as the first item of the
Mechanism's `output_values <PredictionErrorMechanism.output_values>` attribute.

.. _PredictionErrorMechanism_Example:

Example
-------

.. _PredictionErrorMechanism_Default_Input_Value_Example:

*Formatting InputPort values*

The **default_variable** argument can be used to specify a particular format
for the SAMPLE and/or TARGET InputPorts of a PredictionErrorMechanism. This
can be useful when one or both of these differ from the format of the
OutputPort(s) specified in the **sample** and **target** arguments. For
example, for `Temporal Difference Learning <TDLearning>`, a
PredictionErrorMechanism is used to compare the predicted reward from the
sample with the true reward (the target). In the example below, the sample
Mechanism is a `TransferMechanism` that uses the `Linear` function to output
the sample values. Because the output is a vector, specifying it as the
PredictionErrorMechanism's **sample** argument will generate a corresponding
InputPort with a vector as its value. This should match the reward
signal specified in the PredictionErrorMechanism's **target** argument, the
value of which is a vector of the same length as the output of sample.

    >>> import psyneulink as pnl
    >>> sample_mech = pnl.TransferMechanism(input_shapes=5,
    ...                                     function=pnl.Linear())
    >>> reward_mech = pnl.TransferMechanism(input_shapes=5)
    >>> prediction_error_mech = pnl.PredictionErrorMechanism(sample=sample_mech,
    ...                                                      target=reward_mech)

Note that ``sample_mech`` is specified to take an array of length 5 as its
input, and therefore generate one of the same length as its `primary output
<OutputPort_Primary>`. Since it is assigned as the **sample** of the
PredictionErrorMechanism, by default this will create a *SAMPLE* InputPort of
length 5, that will match the length of the *TARGET* InputPort.

Currently the default method of implementing temporal difference learning in
PsyNeuLink requires the values of *SAMPLE* and *TARGET* to be provided as an
array representing a full time series as an experiment. See
`MontagueDayanSejnowski.py` in the Scripts folder for an example.

.. _PredictionErrorMechanism_Class_Reference

Class Reference
---------------

"""
from psyneulink._typing import Iterable

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Union

from psyneulink.core.components.functions.nonstateful.transformfunctions import PredictionErrorDeltaFunction
from psyneulink.core.components.mechanisms.mechanism import Mechanism_Base
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.keywords import PREDICTION_ERROR_MECHANISM, SAMPLE, TARGET
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel, PREFERENCE_SET_NAME
from psyneulink.core.globals.utilities import NumericCollections
from psyneulink.library.components.mechanisms.processing.objective.comparatormechanism import ComparatorMechanism, ComparatorMechanismError

__all__ = [
    'PredictionErrorMechanism',
    'PredictionErrorMechanismError'
]


class PredictionErrorMechanismError(ComparatorMechanismError):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class PredictionErrorMechanism(ComparatorMechanism):
    """
    PredictionErrorMechanism(                                \
        sample,                                              \
        target,                                              \
        function=PredictionErrorDeltaFunction,               \
        output_ports=[OUTCOME],                             \
        params=None,                                         \
        name=None,                                           \
        prefs=None)

    Subclass of ComparatorMechanism that calculates the prediction error between the predicted reward and the target.
    See `ComparatorMechanism <ComparatorMechanism_Class_Reference>` for additional arguments and attributes.

    Arguments
    ---------

    sample : OutputPort, Mechanism_Base, dict, number, or str
        specifies the *SAMPLE* InputPort, that is evaluated by the `function <PredictionErrorMechanism.function>`.

    target : OutputPort, Mechanism_Base, dict, number, or str
        specifies the *TARGET* InputPort used by the function to evaluate `sample<PredictionErrorMechanism.sample>`.

    function : TransformFunction, ObjectiveFunction, function, or method : default PredictionErrorDeltaFunction
        the function used to evaluate the SAMPLE and TARGET inputs.

    learning_rate : Number : default 0.3
        controls the weight of later timesteps compared to earlier ones;  higher rates weight later timesteps more
        heavily than previous ones.

    Attributes
    ----------

    sample : OutputPort, Mechanism_Base, dict, number, or str
        the *SAMPLE* `InputPort`, the `value <InputPort.value>` of which will be evaluated by the function.

    target : OutputPort, Mechanism_Base, dict, number, or str
        the *TARGET* `InputPort`, the `value <InputPort.value>` of which will be used to evaluate `sample
        <PredictionErrorMechanism.sample>`.

    function : TransformFunction, ObjectiveFunction, Function, or method : default PredictionErrorDeltaFunction
        the function used to evaluate the sample and target inputs.

    output_ports : str, Iterable : default OUTCOME
        by default, contains only the *OUTCOME* (`primary <OutputPort_Primary>`) `OutputPort` of the
        PredictionErrorMechanism.

    learning_rate : Number : default 0.3
        controls the weight of later timesteps compared to earlier ones; higher rates weight later timesteps more
        heavily than previous ones.

    """
    componentType = PREDICTION_ERROR_MECHANISM

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    classPreferences = {
        PREFERENCE_SET_NAME: 'PredictionErrorMechanismCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)
    }

    class Parameters(ComparatorMechanism.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <PredictionErrorMechanism.variable>`

                    :default value: None
                    :type:
                    :read only: True

                function
                    see `function <PredictionErrorMechanism.function>`

                    :default value: `PredictionErrorDeltaFunction`
                    :type: `Function`

                learning_rate
                    see `learning_rate <PredictionErrorMechanism.learning_rate>`

                    :default value: 0.3
                    :type: ``float``
        """

        variable = Parameter(None, read_only=True, pnl_internal=True, constructor_argument='default_variable')
        learning_rate = Parameter(0.3, modulable=True)
        function = Parameter(PredictionErrorDeltaFunction, stateful=False, loggable=False)
        sample = None
        target = None

    @check_user_specified
    @beartype
    def __init__(self,
                 sample: Optional[Union[OutputPort, Mechanism_Base, dict, NumericCollections, str]] = None,
                 target: Optional[Union[OutputPort, Mechanism_Base, dict, NumericCollections, str]] = None,
                 function=None,
                 output_ports: Optional[Union[str, Iterable]] = None,
                 learning_rate: Optional[NumericCollections] = None,
                 params=None,
                 name=None,
                 prefs:  Optional[ValidPrefSet] = None,
                 **kwargs
                 ):

        super().__init__(
            sample=sample,
            target=target,
            function=function,
            output_ports=output_ports,
            learning_rate=learning_rate,
            params=params,
            name=name,
            prefs=prefs,
            **kwargs
        )

    def _parse_function_variable(self, variable, context=None):
        # TODO: update to take sample/reward from variable
        # sample = x(t) in Montague on first run, V(t) on subsequent runs
        sample = self.input_ports[SAMPLE].parameters.value._get(context)
        reward = self.input_ports[TARGET].parameters.value._get(context)

        return np.array([sample, reward])

    def _execute(self, variable=None, context=None, runtime_params=None):
        delta = super()._execute(variable=variable, context=context, runtime_params=runtime_params)
        delta = delta[0][1:]
        delta = np.append(delta, 0)
        return delta



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# TODO:

# FIX: TERMINATION CONDITION IS GETTING TRIGGED AFTER 1st TRIAL

# FOR INPUT NODES: scheduler.add_condition(A, BeforeNCalls(A,1)
# Termination: AfterNCalls(Ctl,2)

"""
QUESTIONS:

NOTES:
    *MUST* run Experience before Predict, as the latter requires retrieved_reward to be non-zero
           (from last trial of Experience) in order to know to encode the next state (see control policy)

**Overview**
------------

This implements a model of...

The model is an example of...

The script contains methods to construct, train, and run the model, and analyze the results of its execution:

* `construct_model <EGO.construct_model>`:
  takes as arguments parameters used to construct the model;  for convenience, defaults are defined below,
  (under "Construction parameters")

* `train_network <EGO.train_network>`:
  takes as arguments the feedforward neural network Composition (FFN_COMPOSITION) and number of epochs to train.
  Note: learning_rate is set at construction (can specify using LEARNING_RATE under "Training parameters" below).

* `run_model <EGO.run_model>`:
  takes as arguments the drift rate in the temporal context vector to be applied on each trial,
  and the number of trials to execute, as well as reporting and animation specifications
  (see "Execution parameters").

* `analyze_results <EGO.analyze_results>`:
  takes as arguments the results of executing the model, and optionally a number of trials and EGO_level to analyze;
  returns d-prime statistics and plots results for different conditions at each EGO_level executed.


**The Model**
-------------

The model is comprised of...

.. _EGO_Fig:

.. figure:: _static/N-Back_Model_movie.gif
   :align: left
   :alt: N-Back Model Animation

.. _EGO_model_composition:

*EGO_model Composition*
~~~~~~~~~~~~~~~~~~~~~~~~~

This is comprised of three input Mechanisms, and the nested `ffn <EGO_ffn_composition>` `Composition`.

.. _EGO_ffn_composition:

*FFN Composition*
~~~~~~~~~~~~~~~~~

The temporal context is provided by a randomly drifting high dimensional vector that maintains a constant norm (i.e.,
drifts on a sphere).  The FFN is trained, given an n-back level of *n*, to identify when the current stimulus matches
one stored in EM with a temporal context vector that differs by an amount corresponding to *n* time steps of drift.
During n-back performance, the model encodes the current stimulus and temporal context, retrieves an item from EM
that matches the current stimulus, weighted by the similarity of its temporal context vector (i.e., most recent), and
then uses the FFN to evaluate whether it is an n-back match.  The model responds "match" if the FFN detects a match;
otherwise, it either uses the current stimulus and temporal context to retrieve another sample from EM and repeat the
evaluation or, with a fixed probability (hazard rate), it responds "non-match".

The ffn Composition is trained using the train_network() method


**Construction and Execution**
------------------------------

.. _EGO_settings:

*Settings*
~~~~~~~~~~

The default parameters are ones that have been fit to empirical data concerning human performance
(taken from `Kane et al., 2007 <https://psycnet.apa.org/record/2007-06096-010?doi=1>`_).

See "Settings for running the script" to specify whether the model is trained and/or executed when the script is run,
and whether a graphic display of the network is generated when it is constructed.

.. _EGO_stimuli:

*Stimuli*
~~~~~~~~~

Sequences of stimuli are constructed either using `SweetPea <https://sites.google.com/view/sweetpea-ai?pli=1>`_
(using the script in stim/SweetPea) or replicate those used in the study by `Kane et al.,
2007 <https://psycnet.apa.org/record/2007-06096-010?doi=1>`_ (from stimulus files in stim/Kane_et_al).

    .. note::
       Use of SweetPea for stimulus generation requires it be installed::
       >> pip install sweetpea


.. _EGO_training:

*Training*
~~~~~~~~~~

MORE HERE

.. _EGO_execution:

*Execution*
~~~~~~~~~~~

MORE HERE

.. _EGO_methods_reference:

**Methods Reference**
---------------------


"""

import numpy as np
from enum import IntEnum

from psyneulink import *
from psyneulink._typing import Union, Literal
from psyneulink.core.scheduling.condition import Any, And, AllHaveRun, AtRunStart

# Settings for running script:

NUM_EXP_SEQS = 5               # Number of sequences to run in EXPERIENCE Phase (includes baseline + revaluation)
NUM_PRED_TRIALS = 10           # Number of trials (ROLL OUTS) to run in PREDICTION Phase

CONSTRUCT_MODEL = True                 # THIS MUST BE SET TO True to run the script
DISPLAY_MODEL = (                      # Only one of the following can be uncommented:
    # None                             # suppress display of model
    {}                               # show simple visual display of model
    # {'show_node_structure': True}    # show detailed view of node structures and projections
)
# RUN_MODEL = True                       # True => run the model
RUN_MODEL = False                      # False => don't run the model
ANALYZE_RESULTS = False                # True => output analysis of results of run
REPORT_OUTPUT = ReportOutput.FULL     # Sets console output during run [ReportOutput.ON, .TERSE OR .FULL]
REPORT_PROGRESS = ReportProgress.OFF   # Sets console progress bar during run
PRINT_RESULTS = False                  # print model.results after execution
ANIMATE = False # {UNIT:EXECUTION_SET} # Specifies whether to generate animation of execution


#region   PARAMETERS
# ======================================================================================================================
#                                                   PARAMETERS
# ======================================================================================================================

# PyTorch Version Parameters:
model_params = dict(
    n_participants=58,
    n_simulations = 100, # number of rollouts per participant
    n_steps = 3, # number of steps per rollout
    state_d = 7, # length of state vector
    context_d = 7, # length of context vector
    time_d = 25, # length of time vector
    self_excitation = .25, # rate at which old context is carried over to new context
    input_weight = .5, # rate at which state is integrated into new context
    retrieved_context_weight = .25, # rate at which context retrieved from EM is integrated into new context
    time_noise=.01,# noise std for time integrator (drift is set to 0)
    state_weight = .5, # weight of the state used during memory retrieval
    context_weight = .3, # weight of the context used during memory retrieval
    time_weight = .2, # weight of the time used during memory retrieval
    temperature = .05 # temperature of the softmax used during memory retrieval (smaller means more argmax-like
)

# Fixed (structural) parameters:

# Names:
MODEL_NAME = "EGO Model"
TASK_INPUT_LAYER_NAME = "TASK"
STATE_INPUT_LAYER_NAME = "STATE"
TIME_INPUT_LAYER_NAME = "TIME"
ATTEND_EXTERNAL_LAYER_NAME = "ATTEND\nEXTERNAL\nSTATE"
ATTEND_MEMORY_LAYER_NAME = "ATTEND\nMEMORY\nOF STATE"
CONTROL_LAYER_NAME = "CONTROL"
CONTEXT_LAYER_NAME = 'CONTEXT'
REWARD_INPUT_LAYER_NAME = "REWARD"
# RETRIEVED_TIME_NAME = "RETRIEVED\nTIME"
RETRIEVED_TIME_NAME = "RETRIEVED_TIME"  # NAME OF EM OUTPUT PORT FOR TIME
# RETRIEVED_REWARD_NAME = "RETRIEVED\nREWARD"
RETRIEVED_REWARD_NAME = "RETRIEVED REWARD"
EM_NAME = "EM"
DECISION_LAYER_NAME = "DECISION"
RESPONSE_LAYER_NAME = "RESPONSE"


Task = IntEnum('Task', ['EXPERIENCE', 'PREDICT'],
               start=0)

EMFieldsIndex = IntEnum('EMFields',
                        ['STATE',
                         'TIME',
                         'CONTEXT',
                         'REWARD'],
                        start=0)

StateFeatureIndex = IntEnum('StateFeatureIndex',
                            ['TASK',
                             'REWARD'],
                            start=0)

ControlSignalIndex = IntEnum('ControlSignalIndex',
                             ['ATTEND_EXTERNAL',
                              'ATTEND_MEMORY',
                              'EM_FIELD_WEIGHTS',
                              'STORAGE_PROB',
                              'DECISION_GATE'],
                             start=0)

# CONSTRUCTION PARAMETERS

# Layer sizes:
TASK_SIZE = 1                         # length of task vector
STATE_SIZE = model_params['state_d']  # length of state vector
TIME_SIZE = model_params['time_d']    # length of time vector
REWARD_SIZE = 1                       # length of reward vector
DECISION_SIZE = 1                     # length of decision vector

# Context processing:
STATE_WEIGHT = model_params['input_weight']  # rate at which external vs. memory state are integrated in context_layer
CONTEXT_INTEGRATION_RATE = model_params['retrieved_context_weight']  # rate at which retrieved context (from EM)
                                                                     # is integrated into context_layer
assert (model_params['retrieved_context_weight'] + STATE_WEIGHT + CONTEXT_INTEGRATION_RATE) == 1,\
    (f"Sum of STATE_WEIGHT ({STATE_WEIGHT}), CONTEXT_INTEGRATION_RATE ({CONTEXT_INTEGRATION_RATE}), "
     f"and RETRIEVED_CONTEXT_WEIGHT ({model_params['retrieved_context_weight']}) must equal 1")

# EM retrieval
STATE_RETRIEVAL_WEIGHT = model_params['state_weight']     # weight of state field in retrieval from EM
TIME_RETRIEVAL_WEIGHT = model_params['time_weight']       # weight of time field in retrieval from EM
CONTEXT_RETRIEVAL_WEIGHT = model_params['context_weight'] # weight of context field in retrieval from EM
REWARD_RETRIEVAL_WEIGHT = 0                               # weight of reward field in retrieval from EM
RETRIEVAL_SOFTMAX_GAIN = 1/model_params['temperature']    # gain on softmax retrieval function
# RETRIEVAL_HAZARD_RATE = 0.04   # rate of re=sampling of em following non-match determination in a pass through ffn

RANDOM_WEIGHTS_INITIALIZATION=RandomMatrix(center=0.0, range=0.1)  # Matrix spec used to initialize all Projections

#endregion

#region ENVIRONMENT
# ======================================================================================================================
#                                                   ENVIRONMENT
# ======================================================================================================================

# Temporal context vector generation as input to time_input_layer of model
TIME_DRIFT_RATE = 0.1          # noise used by DriftOnASphereIntegrator (function of Context mech)
TIME_DRIFT_NOISE = 0.0         # noise used by DriftOnASphereIntegrator (function of Context mech)
time_fct = DriftOnASphereIntegrator(initializer=np.random.random(TIME_SIZE - 1),
                                    noise=TIME_DRIFT_NOISE,
                                    dimension=TIME_SIZE)
# Task environment:
NUM_STIM_PER_SEQ = model_params['n_steps'] # number of stimuli in a sequence
NUM_BASELINE_SEQS = NUM_EXP_SEQS     # num trials for Task.EXPERIENCE (passive encoding into EM) BEFORE revaluation
NUM_REVALUATION_SEQS = NUM_EXP_SEQS  # num trials for Task.EXPERIENCE (passive encoding into EM) AFTER revaluation
NUM_EXPERIENCE_SEQS = NUM_BASELINE_SEQS + NUM_REVALUATION_SEQS  # total number of trials for Task.EXPERIENCE
TOTAL_NUM_EXPERIENCE_STIMS = (NUM_BASELINE_SEQS * NUM_STIM_PER_SEQ) + (NUM_REVALUATION_SEQS * (NUM_STIM_PER_SEQ-1))
NUM_ROLL_OUTS = NUM_PRED_TRIALS    # number of times to roll out each sequence in Task.PREDICT

STIM_SEQS = [list(range(1,NUM_STIM_PER_SEQ*2,2)),
            list(range(2,NUM_STIM_PER_SEQ*2+1,2))]
REWARD_VALS = [10,1]         # reward values for seq_0 and seq_1, respectively
RATIO = 1                    # ratio of seq_0 to seq_1 in EXPERIENCE phase (ratio / 1 + ratio)
SAMPLING_TYPE = 'random'     # 'random' or 'alternating' in EXPERIENCE phase
PREDICT_SEQ_TYPE = 'blocked' # 'blocked' or 'alternating' in PREDICT phase

def get_states(state_size=STATE_SIZE):
    state_reps = np.eye(state_size)
    return state_reps

def build_experience_inputs(state_size:int=STATE_SIZE,
                            time_drift_rate:float=TIME_DRIFT_RATE,
                            num_baseline_seqs:int=NUM_BASELINE_SEQS,
                            num_revaluation_seqs:int=NUM_REVALUATION_SEQS,
                            reward_vals:list=REWARD_VALS,
                            sampling_type:Literal[Union['random','alternating']]=SAMPLING_TYPE,
                            ratio:int=RATIO,
                            stim_seqs:list=STIM_SEQS)->tuple:
    """Build inputs for full sequence of trials (with one stim per trial) for EGO MDP model
    Return tuple in which each item is list of all trials for a layer of the model: (time, task, state, reward)
    """
    assert isinstance(ratio,int),f"ratio ({ratio}) must be an integer"
    assert sampling_type in ['random','alternating'], f"Sampling type must be 'random' or 'alternating'"

    def gen_baseline_states_and_rewards(state_size:int=state_size,
                                        stim_seqs:list=stim_seqs,
                                        reward_vals:list=reward_vals,
                                        num_seqs:int=num_baseline_seqs,
                                        sampling_type:Literal[Union['random','alternating']]=sampling_type,
                                        ratio:int=ratio,
                                        )->tuple:
        """Generate states and rewards for reward revaluation phase of Experiment 1
        Return tuple with one-hot representations of (states, rewards, length of a single sequence)
        """
        # Generate one-hots
        state_reps = get_states(state_size)

        # Generate sequence of states
        visited_states, rewards = [], []
        seq_len = len(stim_seqs[0])
        for i in range(num_seqs):
            seq_0 = np.random.random() < (ratio / (ratio + 1)) if sampling_type == 'random' else i % (ratio + 1)
            if seq_0:
                visited_states.extend(stim_seqs[0])
                rewards.extend([0] * (seq_len - 1) + [reward_vals[0]])
            else:
                visited_states.extend(stim_seqs[1])
                rewards.extend([0] * (seq_len - 1) + [reward_vals[1]])

        # Pick one-hots corresponding to each state
        visited_states = state_reps[visited_states]
        rewards = np.array(rewards)

        return visited_states, rewards, seq_len

    def gen_reward_revaluation_states_and_reward(state_size:int=STATE_SIZE,
                                                 stim_seqs:list=stim_seqs,
                                                 reward_vals:list=reward_vals,
                                                 num_seqs:int=num_revaluation_seqs,
                                                 sampling_type:Literal[Union['random','alternating']]=sampling_type,
                                                 ratio:int=ratio,
                                                 )->tuple:
        """Generate states and rewards for reward revaluation phase of Experiment 1
        Return tuple with one-hot representations of (states, rewards, length of a single sequence)
        """

        # Generate one-hots
        state_reps = get_states(state_size)

        # Generate sequence of states
        visited_states, rewards = [], []
        seq_len = len(stim_seqs[0][1:])
        for trial_idx in range(num_seqs):
            seq_0 = np.random.random() < (ratio / (ratio + 1)) if sampling_type == 'random' else trial_idx % (ratio + 1)
            if seq_0:
                visited_states.extend(stim_seqs[0][1:])
                rewards.extend([0] * (seq_len - 1) + [reward_vals[0]])
            else:
                visited_states.extend(stim_seqs[1][1:])
                rewards.extend([0] * (seq_len - 1) + [reward_vals[1]])

        # Pick one-hots corresponding to each state
        visited_states = state_reps[visited_states]
        rewards = np.array(rewards)

        return visited_states, rewards, seq_len

    # Get sequences of states and rewards for baseline trials
    baseline_states, baseline_rewards, num_stim_per_baseline_seq = \
        (gen_baseline_states_and_rewards(state_size=state_size,
                                         stim_seqs=stim_seqs,
                                         reward_vals=reward_vals,
                                         num_seqs=num_baseline_seqs,
                                         sampling_type=sampling_type,
                                         ratio=ratio))

    # Get sequences of states and rewards for reward revaluation trials
    reward_revaluation_states, reward_revaluation_rewards, num_stim_per_revaluation_seq = \
        (gen_reward_revaluation_states_and_reward(state_size=state_size,
                                                  stim_seqs=stim_seqs,
                                                  reward_vals=reward_vals,
                                                  num_seqs=num_revaluation_seqs,
                                                  sampling_type=sampling_type,
                                                  ratio=ratio))

    states = np.concatenate((baseline_states, reward_revaluation_states))
    rewards = np.concatenate((baseline_rewards, reward_revaluation_rewards))

    # Get sequences of task and time inputs
    num_trials = (num_baseline_seqs * num_stim_per_baseline_seq
                             + num_revaluation_seqs * num_stim_per_revaluation_seq)
    tasks = np.full(num_trials, Task.EXPERIENCE.value)
    times = np.array([time_fct(time_drift_rate) for i in range(num_trials)])

    assert len(times) == num_trials
    assert len(tasks) == num_trials
    assert len(states) == num_trials
    assert len(rewards) == num_trials

    return times, tasks, states, rewards

def build_prediction_inputs(state_size:int=STATE_SIZE,
                            time_drift_rate:float=TIME_DRIFT_RATE,
                            num_roll_outs_per_stim:int=int(NUM_ROLL_OUTS / 2),
                            stim_seqs:list=STIM_SEQS,
                            reward_vals:list=REWARD_VALS,
                            seq_type:Literal[Union['blocked','alternate']]=PREDICT_SEQ_TYPE)->tuple:

    # Get stimulus sequences
    num_trials = int(num_roll_outs_per_stim * 2)
    state_reps = get_states(state_size)
    test_stims = [stim_seqs[0][0],stim_seqs[1][0]]
    stim_seq = []
    rewards = []
    for i in range(num_trials):
        seq = int(i * 2 / num_trials) if seq_type == 'blocked' else i % 2
        stim_seq.extend([test_stims[seq]])
        rewards.extend([reward_vals[seq]])

    # Generate time, task and state inputs
    times = np.array([time_fct(time_drift_rate) for i in range(num_trials)])
    tasks = np.full(num_trials, Task.PREDICT.value)
    states =  state_reps[stim_seq]

    return times, tasks, states, np.array(rewards)


#endregion

#region   MODEL
# ======================================================================================================================
#                                                      MODEL
# ======================================================================================================================

def construct_model(model_name:str=MODEL_NAME,

                    # Inputs:
                    task_input_name:str=TASK_INPUT_LAYER_NAME,
                    task_size:int=1,
                    state_input_name:str=STATE_INPUT_LAYER_NAME,
                    state_size:int=STATE_SIZE,
                    time_input_name:str=TIME_INPUT_LAYER_NAME,
                    time_size:int=TIME_SIZE,
                    reward_input_name = REWARD_INPUT_LAYER_NAME,
                    reward_size:int=REWARD_SIZE,

                    # Context processing:
                    # attention_layer_name=ATTENTION_LAYER_NAME,
                    attend_external_layer_name=ATTEND_EXTERNAL_LAYER_NAME,
                    attend_memory_layer_name=ATTEND_MEMORY_LAYER_NAME,
                    attentional_control_name=CONTROL_LAYER_NAME,
                    context_name:str=CONTEXT_LAYER_NAME,
                    state_weight:Union[float,int]=STATE_WEIGHT,
                    context_integration_rate:Union[float,int]=CONTEXT_INTEGRATION_RATE,

                    # EM:
                    em_name:str=EM_NAME,
                    retrieval_softmax_gain=RETRIEVAL_SOFTMAX_GAIN,
                    state_retrieval_weight:Union[float,int]=STATE_RETRIEVAL_WEIGHT,
                    time_retrieval_weight:Union[float,int]=TIME_RETRIEVAL_WEIGHT,
                    context_retrieval_weight:Union[float,int]=CONTEXT_RETRIEVAL_WEIGHT,
                    reward_retrieval_weight:Union[float,int]=REWARD_RETRIEVAL_WEIGHT,
                    retrieved_reward_name:str=RETRIEVED_REWARD_NAME,

                    # Output / decision processing:
                    decision_layer_name:str=DECISION_LAYER_NAME,

                    )->Composition:

    # Apportionment of contributions of state (actual or em) vs. context (em) to context_layer integration:

    # state input (EXPERIENCE) -\
    #                            --> state_weight -------\
    # state from em (PREDICT)---/                         -> * (context_integration_rate) -----\
    #                          /-----> context_weight ---/                                      --> context
    # context from em --------/      (=1- state_weight)                                        /
    #                                                    /---> 1 - context_integration_rate --/
    # context from prev. cycle -------------------------/

    assert 0 <= context_integration_rate <= 1,\
        f"context_retrieval_weight must be a number from 0 to 1"
    assert 0 <= state_weight <= 1,\
        f"context_retrieval_weight must be a number from 0 to 1"
    context_weight = 1 - state_weight
    state_weight *= context_integration_rate
    context_weight *= context_integration_rate

    # ----------------------------------------------------------------------------------------------------------------
    # -------------------------------------------------  Mechanisms  -------------------------------------------------
    # ----------------------------------------------------------------------------------------------------------------

    task_input_layer = ProcessingMechanism(name=task_input_name, input_shapes=task_size)
    state_input_layer = ProcessingMechanism(name=state_input_name, input_shapes=state_size)
    time_input_layer = ProcessingMechanism(name=time_input_name, input_shapes=time_size)
    reward_input_layer = ProcessingMechanism(name=reward_input_name, input_shapes=reward_size)
    attend_external_layer = ProcessingMechanism(name=attend_external_layer_name, input_shapes=state_size)
    attend_memory_layer = ProcessingMechanism(name=attend_memory_layer_name, input_shapes=state_size)
    retrieved_reward_layer = TransferMechanism(name=retrieved_reward_name, input_shapes=reward_size)
    context_layer = RecurrentTransferMechanism(name=context_name,
                                               input_shapes=state_size,
                                               auto=1-context_integration_rate,
                                               hetero=0.0)
    em = EpisodicMemoryMechanism(name=em_name,
                                 default_variable=[[0] * state_size,   # state
                                                   [0] * time_size,    # time
                                                   [0] * state_size,   # context
                                                   [0] * reward_size], # reward
                                 input_ports=[{NAME:state_input_name, INPUT_SHAPES:state_size},
                                              {NAME:time_input_name, INPUT_SHAPES:time_size},
                                              {NAME:context_name, INPUT_SHAPES:state_size},
                                              {NAME:reward_input_name, INPUT_SHAPES:reward_size}],
                                 function=ContentAddressableMemory(
                                     # selection_function=SoftMax(gain=retrieval_softmax_gain),
                                     distance_field_weights=[state_retrieval_weight,
                                                             time_retrieval_weight,
                                                             context_retrieval_weight,
                                                             reward_retrieval_weight]))
    # em.output_ports[RETRIEVED_TIME_NAME].parameters.require_projection_in_composition.set(False, override=True)

    decision_layer = DDM(function=DriftDiffusionIntegrator(noise=0.0, rate=1.0),
                         execute_until_finished=False,
                         reset_stateful_function_when=AtRunStart(),
                         name=decision_layer_name)


    # ----------------------------------------------------------------------------------------------------------------
    # -------------------------------------------------  Control  ----------------------------------------------------
    # ----------------------------------------------------------------------------------------------------------------

    tot = NUM_ROLL_OUTS * NUM_STIM_PER_SEQ
    #                +-------------------------------------------------------------------+
    #                |    KEYS    |                       VALUES                         |
    #                |   state    |                 control_allocation                   |
    #                |  features  |                                                      |
    #                +-------------------------------------------------------------------+
    #                 TASK REWARD | ACTUAL EM | TIME STATE CONTEXT REWARD | STORE | DDM
    control_policy = [[[0], [-1],    [1], [0],  [1,    1,     1,      0],    [1],   [0]], # EXPERIENCE
                      [[1],  [0],    [0], [0],  [1,    0,     1,      0],    [0],   [1]], # PREDICT - ROLLOUT
                      [[1],  [1],    [1], [0],  [1,    1,     1,      0],    [0],   [1]]] # PREDICT - REWARD/ENCODE NEXT
    control_em = ContentAddressableMemory(initializer=control_policy,
                                          distance_function=Distance(metric=EUCLIDEAN),
                                          distance_field_weights=[1, 1, 0, 0, 0, 0, 0],
                                          storage_prob=0.0)
    control_em.name = 'CONTROL EM'
    num_keys = len(StateFeatureIndex)
    num_vals = len(ControlSignalIndex)
    num_fields = num_keys + num_vals

    def control_function(variable, context=None, **kwargs):
        """Use by control_layer to govern EM storage & retrieval, access to context, and gating decision

        - Use control_layer variable ("state_features" per ocm) to get control_allocation from control_em:
            - for Task EXPERIENCE, set distance_field_weights to [1] + [0] * 9
            - for Task PREDICT, set distance_field_weights to [1] * 4 + [0] * 6
        - Use retrieved values from control_em to set control_signals for control_layer

        .. table::
           :align: left
           :alt: Player Piano (aka Production System)
           +--------+------------------------------------------------------------------------------------+-----------+
           |        |                                    **CONTROL POLICY**                              |           |
           |        +-----------------------+------------------------------------------------------------+           |
           |        |   "STATE FEATURES"    |                   CONTROL ALLOCATION                       |           |
           |        | (monitor_for_control) +--------------+---------------------------------------------+           |
           |        |                       |  STATE ATTN  |                  EM CONTROL                 |           |
           |        |                       |              +-------------------------------+-------+-----+           |
           |        |                       | ON CURR PASS |      MATCH ON CURR PASS       | STORE | DDM |           |
           |  NUM   |                       | (W/IN TRIAL) |       (field_weights)         |       |     |           |
           | TRIALS |     TASK     REWARD   |  ACTUAL  EM  |  TIME  STATE  CONTEXT REWARD  |       |     |           |
           +--------+-----------------------+--------------+-------------------------------+-------+-----+-----------+
           |   80   |      EXP      ANY     |    1     0   |   1      1       1       0    |   1   |  0  | TRIAL END |
           +--------+-----------------------+--------------+-------------------------------+-------+-----+-----------+
           |        |      PRED      0      |    0     0   |   1      0       1       0    |   0   |  1  |           |
           | #R/O's |      PRED      1      |    1     0   |   1      1       1       0    |   0   |  1  | TRIAL END |
           +--------+-----------------------+--------------+-------------------------------+-------+-----+-----------+
           NOTES:
           - Requires that EXPERIENCE TRIALS be run first, so that
           - STATE ATTN on first PREDICT trial is [1 0] and RETRIEVED_REWARD is [1]
             (since those are carried over from last trial of EXPERIENCE phase)
           - Control signal in response to retrieved_reward has its effect on next trial
                 (since Projection from retrieved_reward -> control is specified as feedback,
                  so that control will execute immediately after input and before EM (in order to control it),
                  but therefore also before retrieve_reward is updated on the current trial)
           - DDM integrates continuously;  could end run on DDM.when_finished instead of fixed number of trials
           - #R/O's = NUM_ROLL_OUTS
        """

        task = variable[StateFeatureIndex.TASK]
        reward = [1] if variable[StateFeatureIndex.REWARD] else [0]
        keys = [task, reward]
        # set values to 0 since they don't matter (should only be retrieving based on keys)
        query = np.array(keys + [[0],[0],[0,0,0,0],[0],[0]], dtype=object)

        # FIX: IF PARAMETERS ARE USED AND CONTEXT IS ADDED TO CALLS BELOW, FAILS COMPLAINING ABOUT NEED FOR SEED
        if task == Task.EXPERIENCE:
            # Set distance_field_weights for EXPERIENCE
            # control_em.parameters.distance_field_weights.set([1] + [0] * (num_fields - 1))
            control_em.distance_field_weights = [1] + [0] * (num_fields - 1)
            # Get control_signals for EXPERIENCE
            control_signals = control_em.execute(query)[num_keys:]

        elif task == Task.PREDICT:
            # Set distance_field_weights for PREDICT
            # control_em.parameters.distance_field_weights.set([1] * num_keys + [0] * num_vals)
            control_em.distance_field_weights = [1] * num_keys + [0] * num_vals
            # Set control_signals for PREDICT
            control_signals = control_em.execute(query)[num_keys:]

        return control_signals

    # Monitored for control
    state_features = [None] * len(StateFeatureIndex)
    state_features[StateFeatureIndex.TASK] = task_input_layer
    state_features[StateFeatureIndex.REWARD] = (retrieved_reward_layer, FEEDBACK)

    # Control signals
    control_signals = [None] * len(ControlSignalIndex)
    control_signals[ControlSignalIndex.ATTEND_EXTERNAL] = (SLOPE, attend_external_layer)
    control_signals[ControlSignalIndex.ATTEND_MEMORY] = (SLOPE, attend_memory_layer)
    control_signals[ControlSignalIndex.EM_FIELD_WEIGHTS] = ('distance_field_weights', em)
    control_signals[ControlSignalIndex.STORAGE_PROB] = (STORAGE_PROB, em)
    control_signals[ControlSignalIndex.DECISION_GATE] = decision_layer.input_port

    control_layer = ControlMechanism(name=attentional_control_name,
                                     monitor_for_control=state_features,
                                     function = control_function,
                                     control=control_signals)
    
    # ----------------------------------------------------------------------------------------------------------------
    # -------------------------------------------------  EGO Composition  --------------------------------------------
    # ----------------------------------------------------------------------------------------------------------------
    
    EGO_comp = Composition(name=model_name,
                           # # Terminate a Task.PREDICT trial after decision_layer executes if a reward is retrieved
                           # termination_processing={
                           #     # TimeScale.TRIAL: And(Condition(lambda: task_input_layer.value == Task.PREDICT),
                           #     #                      Condition(lambda: retrieved_reward_layer.value),
                           #     #                      JustRan(decision_layer))}
                           #     # CRASHES:
                           #     # TimeScale.TRIAL: Any(And(Condition(lambda: task_input_layer.value == Task.EXPERIENCE),
                           #     #                          JustRan(em)),
                           #     #                      And(Condition(lambda: task_input_layer.value == Task.PREDICT),
                           #     #                          Condition(lambda: retrieved_reward_layer.value),
                           #     #                          JustRan(decision_layer)))}
                           #     TimeScale.TRIAL: Any(And(Condition(lambda: task_input_layer.value == Task.EXPERIENCE),
                           #                              AllHaveRun()),
                           #                          And(Condition(lambda: task_input_layer.value == Task.PREDICT),
                           #                              Condition(lambda: retrieved_reward_layer.value),
                           #                              AllHaveRun()))}
                           )

    # Nodes not included in (decision output) Pathway specified above
    EGO_comp.add_nodes([task_input_layer,
                        state_input_layer,
                        time_input_layer,
                        attend_external_layer,
                        attend_memory_layer,
                        context_layer,
                        reward_input_layer,
                        em,
                        retrieved_reward_layer,
                        control_layer,
                        decision_layer
                        ])
    EGO_comp.exclude_node_roles(task_input_layer, NodeRole.OUTPUT)

    # Projections:

    # EM encoding --------------------------------------------------------------------------------
    # state -> em
    EGO_comp.add_projection(MappingProjection(state_input_layer, em.input_ports[STATE_INPUT_LAYER_NAME]))
    # time -> em
    EGO_comp.add_projection(MappingProjection(time_input_layer, em.input_ports[TIME_INPUT_LAYER_NAME]))
    # context -> em
    EGO_comp.add_projection(MappingProjection(context_layer, em.input_ports[CONTEXT_LAYER_NAME]))
    # reward -> em
    EGO_comp.add_projection(MappingProjection(reward_input_layer, em.input_ports[REWARD_INPUT_LAYER_NAME]))

    # Inputs to Context ---------------------------------------------------------------------------
    # actual state -> attend_external_layer
    EGO_comp.add_projection(MappingProjection(state_input_layer, attend_external_layer))
    # retrieved state -> attend_memory_layer
    EGO_comp.add_projection(MappingProjection(em.output_ports[f'RETRIEVED_{STATE_INPUT_LAYER_NAME}'],
                                              attend_memory_layer))
    # attend_external_layer -> context_layer
    EGO_comp.add_projection(MappingProjection(attend_external_layer, context_layer,
                                              matrix=np.eye(STATE_SIZE) * state_weight))
    # attend_memory_layer -> context_layer
    EGO_comp.add_projection(MappingProjection(attend_memory_layer, context_layer,
                                              matrix=np.eye(STATE_SIZE) * state_weight))
    # retrieved context -> context_layer
    EGO_comp.add_projection(MappingProjection(em.output_ports[f'RETRIEVED_{CONTEXT_LAYER_NAME}'], context_layer,
                                              matrix=np.eye(STATE_SIZE) * context_weight))

    # Decision pathway ---------------------------------------------------------------------------
    # retrieved reward -> retrieved reward
    EGO_comp.add_projection(MappingProjection(em.output_ports[f'RETRIEVED_{REWARD_INPUT_LAYER_NAME}'],
                                              retrieved_reward_layer))
    # retrieved reward -> decision layer
    EGO_comp.add_projection(MappingProjection(retrieved_reward_layer, decision_layer))

    # Validate construction
    proj_from_retrieved_reward_to_control = control_layer.input_ports[1].path_afferents[0]
    assert proj_from_retrieved_reward_to_control._feedback == True
    assert proj_from_retrieved_reward_to_control in EGO_comp.feedback_projections # retrieved_reward feedback
    assert context_layer.input_port.path_afferents[0].sender.owner == context_layer # recurrent projection
    assert context_layer.input_port.path_afferents[0].parameters.matrix.get()[0][0] == 1-context_integration_rate
    assert context_layer.input_port.path_afferents[1].sender.owner == attend_external_layer # external state
    assert context_layer.input_port.path_afferents[1].parameters.matrix.get()[0][0] == state_weight
    assert context_layer.input_port.path_afferents[2].sender.owner == attend_memory_layer # memory of state
    assert context_layer.input_port.path_afferents[2].parameters.matrix.get()[0][0] == state_weight
    assert context_layer.input_port.path_afferents[3].sender.owner == em  # memory of context
    assert context_layer.input_port.path_afferents[3].parameters.matrix.get()[0][0] == context_weight
    assert control_layer.input_ports[0].path_afferents[0].sender.owner == task_input_layer
    assert control_layer.input_ports[1].path_afferents[0].sender.owner == retrieved_reward_layer

    return EGO_comp
#endregion


#region SCRIPT EXECUTION
# ======================================================================================================================
#                                                   SCRIPT EXECUTION
# ======================================================================================================================

if __name__ == '__main__':
    model = None

    if CONSTRUCT_MODEL:
        print(f'Constructing {MODEL_NAME}')
        model = construct_model()
        assert 'DEBUGGING BREAK POINT'

    if DISPLAY_MODEL is not None:
        if model:
            model.show_graph(**DISPLAY_MODEL)
        else:
            print("Model not yet constructed")

    if RUN_MODEL:
        experience_inputs = build_experience_inputs(state_size=STATE_SIZE,
                                                    time_drift_rate=TIME_DRIFT_RATE,
                                                    num_baseline_seqs=NUM_BASELINE_SEQS,
                                                    num_revaluation_seqs=NUM_REVALUATION_SEQS,
                                                    reward_vals=REWARD_VALS,
                                                    sampling_type=SAMPLING_TYPE,
                                                    ratio=RATIO,
                                                    stim_seqs=STIM_SEQS)
        input_layers = [TIME_INPUT_LAYER_NAME,
                        TASK_INPUT_LAYER_NAME,
                        STATE_INPUT_LAYER_NAME,
                        REWARD_INPUT_LAYER_NAME]

        # Experience Phase
        print(f"Presenting {model.name} with {TOTAL_NUM_EXPERIENCE_STIMS} EXPERIENCE stimuli")
        model.run(inputs={k: v for k, v in zip(input_layers, experience_inputs)},
                  report_output=REPORT_OUTPUT,
                  report_progress=REPORT_PROGRESS)

        # Prediction Phase
        prediction_inputs = build_prediction_inputs(state_size=STATE_SIZE,
                                                    time_drift_rate=TIME_DRIFT_RATE,
                                                    num_roll_outs_per_stim=int(NUM_ROLL_OUTS / 2),
                                                    stim_seqs=STIM_SEQS,
                                                    reward_vals=REWARD_VALS,
                                                    seq_type=PREDICT_SEQ_TYPE)
        print(f"Running {model.name} for {NUM_ROLL_OUTS} PREDICT (ROLL OUT) trials")
        model.termination_processing = {
            TimeScale.TRIAL: And(Condition(lambda: model.nodes[TASK_INPUT_LAYER_NAME].value == Task.PREDICT),
                                 Condition(lambda: model.nodes[RETRIEVED_REWARD_NAME].value),
                                 # JustRan(model.nodes[DECISION_LAYER_NAME])
                                 AllHaveRun()
                                 )
        }
        model.run(inputs={k: v for k, v in zip(input_layers, prediction_inputs)},
                  report_output=REPORT_OUTPUT,
                  report_progress=REPORT_PROGRESS
                  )

        if PRINT_RESULTS:
            print(f"Predicted reward for last stimulus: {model.results}")
    #endregion


from psyneulink import *
import numpy as np

# Mechanisms
Input = TransferMechanism(name='Input')
reward = TransferMechanism(output_ports=[RESULT, MEAN, VARIANCE],
                               name='reward')
Decision = DDM(function=DriftDiffusionAnalytical(drift_rate=(1.0,
                                                             ControlProjection(function=Linear,
                                                                               control_signal_params={
                                                                                   ALLOCATION_SAMPLES:
                                                                                       np.arange(0.1, 1.01, 0.3)})),
                                                 threshold=(1.0,
                                                            ControlProjection(function=Linear,
                                                                              control_signal_params={
                                                                                  ALLOCATION_SAMPLES:
                                                                                      np.arange(0.1, 1.01, 0.3)})),
                                                 noise=0.5,
                                                 starting_value=0,
                                                 non_decision_time=0.45),
                   output_ports=[DECISION_VARIABLE,
                                RESPONSE_TIME,
                                PROBABILITY_UPPER_THRESHOLD],
                   name='Decision')

comp = Composition(name="evc")
comp.add_node(reward, required_roles=[NodeRole.OUTPUT])
comp.add_node(Decision, required_roles=[NodeRole.OUTPUT])
task_execution_pathway = [Input, IDENTITY_MATRIX, Decision]
comp.add_linear_processing_pathway(task_execution_pathway)

comp.add_model_based_optimizer(optimizer=OptimizationControlMechanism(name='OCM',
                                                                      agent_rep=comp,
                                                                      state_features=[Input.input_port, reward.input_port],
                                                                      state_feature_function=AdaptiveIntegrator(rate=0.5),
                                                                      objective_mechanism=ObjectiveMechanism(
                                                                              name='OCM Objective Mechanism',
                                                                              function=LinearCombination(operation=PRODUCT),
                                                                              monitor=[reward,
                                                                                       Decision.output_ports[PROBABILITY_UPPER_THRESHOLD],
                                                                                       (Decision.output_ports[RESPONSE_TIME], -1, 1)]),
                                                                      function=GridSearch(),
                                                                      control_signals=[("drift_rate", Decision),
                                                                                       ("threshold", Decision)])
                               )

comp.enable_model_based_optimizer = True

stim_list_dict = {
    Input: [0.5, 0.123],
    reward: [20, 20]
}

comp.show_graph(show_controller=True, show_node_structure=ALL)
# comp.show_graph(show_controller=True)
# comp.run(inputs=stim_list_dict)
